/*=======================================================================================================*/
/*      Distributed Speech Recognition                                                                   */
/*      ETSI ES 202 050 Advanced Front-End Feature Extraction Algorithm & Compression Algorithm &        */
/*      ETSI ES 202 212 Extended Advanced Front-End Feature Extraction Algorithm & Compression Algorithm */  
/*      Bit exact C-language software implementation using ETSI basic operations library                 */
/*      3GPP TS 26.243 v 6.1.0 November 2004                                                                  */
/*      ANSI C code for the Fixed-Point Distributed Speech Recognition Extended Advanced Front-end       */
/*=======================================================================================================*/
/*---------------------------------------------------------------------------
 *
 * FILE NAME: BufferIn32.c
 * PURPOSE: Signal buffer.
 *
 *---------------------------------------------------------------------------*/
/*-----------------
 * File Inclusions
 *-----------------*/
#include <stdlib.h>
#include "BufferIn32.h"
#include "typedef.h"
#include "basic_op.h"
#include "count.h"
/*----------------------------------------------------------------------------
 * FUNCTION NAME: BufInAlloc
 *
 * PURPOSE:       Memory allocation of a new buffer. The buffer is filled in 
 *                with zeroes.
 *
 * INPUT:
 *   BufSize      Buffer size
 *
 * OUTPUT:
 *
 *
 * RETURN VALUE:
 *   NULL         In case of memory allocation error
 *   newBuf       Buffer pointer otherwise
 *
 *---------------------------------------------------------------------------*/
extern BufferIn32 *BufIn32Alloc (int BufSize)
{
  BufferIn32 *newBuf = calloc (1, sizeof (BufferIn32));
  if (newBuf == NULL) return NULL;
  newBuf->data = calloc (BufSize, sizeof (newBuf->data[0]));
  if (newBuf->data == NULL) return BufIn32Free (newBuf);
  newBuf->size = BufSize;
  return newBuf;
}

/*----------------------------------------------------------------------------
 * FUNCTION NAME: BufInFree
 *
 * PURPOSE:       Memory free of a buffer.
 *
 * INPUT:
 *   Buf          Pointer to buffer
 *
 * OUTPUT:
 *   none
 *
 * RETURN VALUE:
 *   NULL         In any case, so one can write Buf = BufInFree (Buf);
 *
 *---------------------------------------------------------------------------*/
extern BufferIn32 *BufIn32Free (BufferIn32 *Buf)
{
  if (Buf != NULL)
    {
      free (Buf->data);
      free (Buf);
    }
  return NULL;
}

/*----------------------------------------------------------------------------
 * FUNCTION NAME: BufIn32Clear
 *
 * PURPOSE:       Fill in the buffer with zeroes.
 *
 * INPUT:
 *   Buf          Pointer to buffer
 *
 * OUTPUT:        
 *                Zeroes in Buf->data[]
 *
 * RETURN VALUE:
 *   NULL         In any case, so one can write Buf = BufInFree (Buf) ;
 *
 *---------------------------------------------------------------------------*/
extern void BufIn32Clear (BufferIn32 *Buf)
{
  int i;

  for (i=0 ; i<Buf->size ; i++) 
  {
      Buf->data[i] = 0; move32();
  }
}

/*----------------------------------------------------------------------------
 * FUNCTION NAME: BufIn32Put
 *
 * PURPOSE:       Put data from newData in the buffer. Size of data put in the
 *                buffer is max (newDataSize, Buf->size).
 *
 * INPUT:
 *   Buf          Pointer to buffer
 *   newData      Pointer to new data
 *   newDataSize  Size of new data to put in the buffer
 *
 * OUTPUT:
 *                newData in Buf->data[]
 *
 * RETURN VALUE:
 *   newDataSize  Number of samples of newData put in the buffer
 *
 *---------------------------------------------------------------------------*/
extern int BufIn32Put (BufferIn32 *Buf, X_INT32 *newData, int newDataSize)
{
  int i;

  /* Debug - the line could be protected by #ifdef _DEBUG [..] #endif
     and assert() could also be used */
  if (newDataSize > Buf->size) newDataSize = Buf->size;
  /*----------------
   * shift old data 
   *----------------*/
  for (i=0; i<Buf->size - newDataSize ; i++)
    {
      Buf->data[i] = Buf->data[i + newDataSize]; move32();
    }
  /*----------------------------
   * put new data in the buffer
   *----------------------------*/
  for ( ; i<Buf->size ; i++)
    {
      Buf->data[i] = *newData++; move32();
    }
  return newDataSize;
}

/*----------------------------------------------------------------------------
 * FUNCTION NAME: BufIn32ShiftToPut
 *
 * PURPOSE:       Shift data in the buffer to prepare to put new data in the 
 *                buffer. Return the location where to put new data.
 *
 * INPUT:
 *   Buf          Pointer to buffer
 *   newDataSize  Size of new data to prepare to put in the buffer
 *
 * OUTPUT:
 *                Shift Buf->data[]
 *
 * RETURN VALUE:
 *   NULL         If newDataSize > Buf->size
 *   Buf->data+i  Otherwise a pointer where to insert new data
 *
 *---------------------------------------------------------------------------*/
extern X_INT32 *BufIn32ShiftToPut (BufferIn32 *Buf, int newDataSize)
{
  /* Debug - the line could be protected by #ifdef _DEBUG [..] #endif
     and assert() could also be used */
  if (newDataSize > Buf->size)
    {
      return NULL;
    } 
  else 
    {
      int i;
      /*----------------
       * shift old data 
       *----------------*/
      for (i=0 ; i<Buf->size - newDataSize ; i++)
        {
          Buf->data[i] = Buf->data[i+newDataSize]; move32();
        }
      return Buf->data + i;
    }
}

/*----------------------------------------------------------------------------
 * FUNCTION NAME: BufIn32GetLast
 *
 * PURPOSE:       Output last nbAskSamples of Buf->data[] to dataOutBuf[]
 *
 * INPUT:
 *   Buf          Buffer pointer
 *   nbAskSamples Number of samples to output, i.e. max (nbAskSamples, Buf->size)
 *
 * OUTPUT:
 *                output Buf->data[] in dataOutBuf[]
 *
 * RETURN VALUE:
 *   nbAskSamples Number of samples written in dataOutBuf
 *
 *---------------------------------------------------------------------------*/
extern int BufIn32GetLast (BufferIn32 *Buf, X_INT32 *dataOutBuf, int nbAskSamples)
{
  int i;

  /* Debug - the line could be protected by #ifdef _DEBUG [..] #endif
     and assert() could also be used */
  if (nbAskSamples > Buf->size) nbAskSamples = Buf->size;
  
  /*------------------
   * output last data
   *------------------*/
  for (i=Buf->size - nbAskSamples ; i<Buf->size ; i++)
    {
      *dataOutBuf++ = Buf->data[i]; move32();
    }
  return nbAskSamples;

}

/*----------------------------------------------------------------------------
 * FUNCTION NAME: BufIn32GetPointerToLast
 *
 * PURPOSE:       Return pointer to last nbAskSamples samples of Buf->data[]
 *
 * INPUT:
 *   Buf          Pointer to buffer
 *   nbAskSamples Number of samples to point to
 *
 * OUTPUT:
 *
 * RETURN VALUE:
 *   NULL         If nbAskSamples > Buf->size
 *   pointer      Otherwise pointer to last nbAskSamples samples of Buf->data[]
 *
 *---------------------------------------------------------------------------*/
extern  X_INT32 *BufIn32GetPointerToLast (BufferIn32 *Buf, int nbAskSamples) 
{
  /* Debug - the line could be protected by #ifdef _DEBUG [..] #endif
     and assert() could also be used */
  if (nbAskSamples > Buf->size)
    {
      return NULL;
    }
  else
    {
      return Buf->data + (Buf->size - nbAskSamples);
    }
}
