/*=======================================================================================================*/
/*      Distributed Speech Recognition                                                                   */
/*      ETSI ES 202 050 Advanced Front-End Feature Extraction Algorithm & Compression Algorithm &        */
/*      ETSI ES 202 212 Extended Advanced Front-End Feature Extraction Algorithm & Compression Algorithm */  
/*      Bit exact C-language software implementation using ETSI basic operations library                 */
/*      3GPP TS 26.243 v 6.1.0 November 2004                                                                  */
/*      ANSI C code for the Fixed-Point Distributed Speech Recognition Extended Advanced Front-end       */
/*=======================================================================================================*/

#include "ff4nrFix32_B.h"

#include "typedef.h"
#include "basic_op.h"
#include "count.h"

/*  - If FFT_SHL_1 is defined the signal is multiplied by 2 at the first stage
 *    and the result is divided by 2 at the end (more precision, but saturations can
 *    occur if the input signal amplitude is maximum). */
#define FFT_SHL_1

/* size of sine and cosine tables : Max FFT length / 8 */
#define SIZE_SINE              64


/* prototypes for local fonction */
static void Permut(X_INT32 *b0,X_INT16 M, X_INT16 N) ;
static X_INT16 GetL15(X_INT32 b32);
static X_INT16 GetH15(X_INT32 b32,X_INT16 bl16);
static X_INT32 Mult16x16_16(X_INT16 a16,X_INT16 bl16,X_INT16 bh16) ;
static X_INT32 Mult16x32(X_INT16 a16,X_INT32 b32) ;

#if defined(FFT_SHL_1)
static X_INT32 Add_Mult16x16_16(X_INT16 xa16,X_INT16 xbl16,X_INT16 xbh16, X_INT16 ya16,X_INT16 ybl16,X_INT16 ybh16);
static X_INT32 Sub_Mult16x16_16(X_INT16 xa16,X_INT16 xbl16,X_INT16 xbh16, X_INT16 ya16,X_INT16 ybl16,X_INT16 ybh16);
#endif

extern const X_INT16 tabSin[];
extern const X_INT16 tabCos[];



/*----------------------------------------------------------------------------
 * FUNCTION NAME: ff4NRFix32_B
 *
 * PURPOSE: In place FFT for an INT32 real signal withe size N = 2**M
 *              
 *
 * INPUT:
 *      - b   input real signal
 *      - M   Logarithm in base 2 of FFT length
 *
 * OUTPUT:
 *      - b  : FFT ouput  
 *   b[0] = Re[0]
 *   b[1] = Re[N/2]
 *   b[2] = Re[1]
 *   b[3] = -Im[1]
 *   ...
 *   b[2*k] = Re[k]
 *   b[2*k+1] = -Im[k]
 *
 * WARNING:
 *  - The routines returns the opposite of the imaginary parts.
 *
 * RETURN VALUE: none
 *   
 *
 *---------------------------------------------------------------------------*/

void ff4NRFix32_B(
        int         M,  /* (i)   Logarithm in base 2 of FFT length */
        X_INT32     *b  /* (i/o) int32 signal array                */
)

{
    const static X_INT16 P7 = (X_INT32) (0.70710681 * 32768 + 0.5) ; /*  1 / SQRT(2)  */
    X_INT16   N ,ji,jl,jr,jthet,k ;
    X_INT16   intc ;
    X_INT32   *b0;
    b0 = b ;

/*  Compute N = 2**M */
    N = (X_INT16) shl((X_INT16)1,(X_INT16)M) ;

#if !defined(FFT_SHL_1) 
/* RADIX 2 if necessary */
    test(); logic16();
    if(M & 1) {
        intc = (X_INT16) shr(N,1) ;
        {
            X_INT16 k ;
            for(k=0;k<intc;k++) {
                X_INT32 T0 ;
                T0 = L_add(b0[k],b0[k+intc]) ;
                b0[k+intc] = L_sub(b0[k],b0[k+intc]) ; move32();
                b0[k] = T0 ;   move32();
            }
        }
    } else {
        intc = N ; move16();
    }
#else /* FFT_SHL_1 */
    test(); logic16();
    if(M & 1) { /* RADIX 2 if necessary */
        intc = (X_INT16) shr(N,1) ; move16();
        {
            X_INT16 k  ;
            for(k=0;k<intc;k++) {
                X_INT32 T0,T1 ;
                T0 = L_shl(b0[k],1) ; 
                T1 = L_shl(b0[k+intc],1) ; 
                b0[k+intc] = L_sub(T0,T1) ; move32();
                b0[k] = L_add(T0,T1) ;   move32();
            }
        }
    } else { /* first RADIX 4 stage */
        X_INT32 * b1 ;
        X_INT32 * b2 ;
        X_INT32 * b3 ;
        intc = shr(N,2) ;
        b1 = b0 + intc; move16();
        b2 = b1 + intc; move16();
        b3 = b2 + intc; move16();

        /* Simple radix  BFYR0 */
        for (k = 0 ; k < intc ; k++){
            X_INT32 T0,T1,T2,T3 ;
            T0 = L_shl(b0[k],1) ;
            T1 = L_shl(b1[k],1) ; 
            T2 = L_shl(b2[k],1) ; 
            T3 = L_shl(b3[k],1) ;

            b2[k] = L_sub(T0,T2) ; move32();
            b3[k] = L_sub(T1,T3) ; move32();
            T0 = L_add(T0,T2) ;
            T1 = L_add(T1,T3) ;
            b0[k] = L_add(T0,T1) ; move32();
            b1[k] = L_sub(T0,T1) ; move32();
        }
    }
#endif /* FFT_SHL_1 */

    /* loop for  remaining RADIX 4 */
    test();
#if !defined(FFT_SHL_1)
    while (sub(intc,1) > 0 ) {
#else
    while (sub(intc,4) > 0 ) {
#endif
        X_INT32 * b1 ;
        X_INT32 * b2 ;
        X_INT32 * b3 ;
        intc = shr(intc,2) ;
        b1 = b0 + intc; move16();
        b2 = b1 + intc; move16();
        b3 = b2 + intc; move16();
        /* Simple radix  BFYR0 */
        for (k = 0 ; k < intc ; k++){
            X_INT32 T0,T1 ;
            T0 = L_add(b0[k],b2[k]) ; 
            T1 = L_add(b1[k],b3[k]) ; 
            b2[k] = L_sub(b0[k],b2[k]) ; move32();
            b3[k] = L_sub(b1[k],b3[k]) ; move32();
            b0[k] = L_add(T0,T1) ; move32();
            b1[k] = L_sub(T0,T1) ; move32();
  
        }
        /* BFYI0 if necessary */
        test();
        if ( sub(N,shl(intc,2)) > 0) {
            shl(intc,2); add(shl(intc,2),intc);
            for(k = (X_INT16) (4 * intc) ; k < 5 * intc ; k++) {
                X_INT32 PR,PII ;
                PR = Mult16x32 (P7,L_sub(b1[k],b3[k])); 
                PII = Mult16x32 (P7,L_add(b1[k],b3[k]));
                b3[k] = L_add(b2[k],PII) ; move32();
                b1[k] = L_sub(PII,b2[k]) ; move32();
                b2[k] = L_sub(b0[k],PR) ; move32();
                b0[k] = L_add(b0[k],PR) ; move32();
            }
        }
        jl = shl(intc,3); 
        ji = add(jl,shl(intc,2)) ; 
        jr = jl ; move16();
        jthet = 1 ; move16();
        /* loop for Radix4 BFY */
        test();
        while (sub( jr, N) < 0 ) {
            X_INT16 j ;
            X_INT16   c1,c2,c3 ;
            X_INT16   s1,s2,s3 ;
            test();
            s1 = tabSin[jthet] ; move16() ;
            c1 = tabCos[jthet] ; move16() ;
            c2 = round(L_msu(L_mult(c1,c1),s1,s1)) ; 
            s2 = round(L_mac(L_mult(s1,c1),c1,s1)) ; 
            c3 = round(L_msu(L_mult(c1,c2),s1,s2)) ; 
            s3 = round(L_mac(L_mult(s1,c2),c1,s2)) ;
            
            k = sub(ji,jr) ;
            for( j = jr ; j < add(jr,intc) ; j++) {
                X_INT32   T0,T1,T2,T3,T4,T5,T6,T7 ;
                X_INT32   TR1,TR2,TR3 ;
                X_INT32   TI1,TI2,TI3 ;
                {
                    X_INT16 l_j, h_j, l_k, h_k ;
                    l_j = GetL15(b1[j]); 
                    h_j = GetH15(b1[j],l_j); 
                    l_k = GetL15(b1[j+k]);  
                    h_k = GetH15(b1[j+k],l_k); 
#if defined(FFT_SHL_1)
                    TR1 = Sub_Mult16x16_16(c1,l_j,h_j,s1,l_k,h_k);
                    TI1 = Add_Mult16x16_16(s1,l_j,h_j,c1,l_k,h_k);		
#else
                    TR1 = L_sub(Mult16x16_16(c1,l_j,h_j) , Mult16x16_16(s1,l_k,h_k));
                    TI1 = L_add(Mult16x16_16(s1,l_j,h_j) , Mult16x16_16(c1,l_k,h_k));
#endif
                    l_j = GetL15(b2[j]); 
                    h_j = GetH15(b2[j],l_j); 
                    l_k = GetL15(b2[j+k]); 
                    h_k = GetH15(b2[j+k],l_k);
#if defined(FFT_SHL_1)
                    TR2 = Sub_Mult16x16_16(c2,l_j,h_j,s2,l_k,h_k);
                    TI2 = Add_Mult16x16_16(s2,l_j,h_j,c2,l_k,h_k);

#else
                    TR2 = L_sub(Mult16x16_16(c2,l_j,h_j) , Mult16x16_16(s2,l_k,h_k));
                    TI2 = L_add(Mult16x16_16(s2,l_j,h_j) , Mult16x16_16(c2,l_k,h_k));

#endif
                    l_j = GetL15(b3[j]);
                    h_j = GetH15(b3[j],l_j);
                    l_k = GetL15(b3[j+k]);
                    h_k = GetH15(b3[j+k],l_k);
#if defined(FFT_SHL_1)
                    TR3 = Sub_Mult16x16_16(c3,l_j,h_j,s3,l_k,h_k);
                    TI3 = Add_Mult16x16_16(s3,l_j,h_j,c3,l_k,h_k);
#else
                    TR3 = L_sub(Mult16x16_16(c3,l_j,h_j) , Mult16x16_16(s3,l_k,h_k));
                    TI3 = L_add(Mult16x16_16(s3,l_j,h_j) , Mult16x16_16(c3,l_k,h_k));
#endif
                }
                T0 = L_add(b0[j] ,   TR2) ;
                T4 = L_add(b0[j+k] , TI2) ;
                T2 = L_sub(b0[j] ,   TR2) ;
                T6 = L_sub(b0[j+k] , TI2) ;

                T1 = L_add(TR1,TR3);
                T5 = L_add(TI1,TI3);
                T3 = L_sub(TR1,TR3);
                T7 = L_sub(TI1,TI3);

                b0[j] =   L_add(T0 , T1) ; move32() ;
                b3[j+k] = L_add(T4 , T5) ; move32() ;
                b2[j+k] = L_sub(T0 , T1) ; move32() ;
                b1[j] =   L_sub(T5 , T4) ; move32() ;
                b2[j] =   L_sub(T2 , T7) ; move32() ;
                b1[j+k] = L_add(T6 , T3) ; move32() ;
                b0[j+k] = L_add(T2 , T7) ; move32() ;
                b3[j] =   L_sub(T3 , T6) ; move32() ;
                
            }
            jthet = add(jthet,1) ;
            jr = add(jr,shl(intc,3)) ;
            ji = sub(ji,shl(intc,3)) ;
            test();
            if ( sub(ji,jl)<0 ) {
                ji = shl(sub(jr,shl(intc,1)),1);
                jl = jr ; move16();
            }
            test(); /* while */
        }
        test(); /* while */
    }
#if defined(FFT_SHL_1)
    /* last pass with divide by 2 */
    test();
    if (sub(intc,1) > 0 ) 
    {
        X_INT32 * b1 ;
        X_INT32 * b2 ;
        X_INT32 * b3 ;
        intc = shr(intc,2) ;
        b1 = b0 + intc; move16();
        b2 = b1 + intc; move16();
        b3 = b2 + intc; move16();

        /* Simple radix  BFYR0 */
        for (k = 0 ; k < intc ; k++){
            X_INT32 T0,T1 ;
            T0 = L_add(b0[k],b2[k]) ; 
            T1 = L_add(b1[k],b3[k]) ; 
            b2[k] = L_shr(L_sub(b0[k],b2[k]),1) ; move32();
            b3[k] = L_shr(L_sub(b1[k],b3[k]),1) ; move32();
            b0[k] = L_shr(L_add(T0,T1),1) ; move32();
            b1[k] = L_shr(L_sub(T0,T1),1) ; move32();
  
        }
        /* BFYI0 if necessary */
        test();
        if ( sub(N,shl(intc,2)) > 0) {
            shl(intc,2); add(shl(intc,2),intc);
            for(k = (X_INT16) (4 * intc) ; k < 5 * intc ; k++) {
                X_INT32 PR,PII ;
                PR = Mult16x32 (P7,L_sub(b1[k],b3[k])); 
                PII = Mult16x32 (P7,L_add(b1[k],b3[k]));
                b3[k] = L_shr(L_add(b2[k],PII),1) ; move32();
                b1[k] = L_shr(L_sub(PII,b2[k]),1) ; move32();
                b2[k] = L_shr(L_sub(b0[k],PR),1) ; move32();
                b0[k] = L_shr(L_add(b0[k],PR),1) ; move32();
            }
        }
        jl = shl(intc,3); 
        ji = add(jl,shl(intc,2)) ; 
        jr = jl ; move16();
        jthet = 1 ; move16();
        /* loop for Radix4 BFY */
        test();
        while (sub( jr, N) < 0 ) {
            X_INT16 j ;
            X_INT16   c1,c2,c3 ;
            X_INT16   s1,s2,s3 ;
            test();
            s1 = tabSin[jthet] ; move16() ;
            c1 = tabCos[jthet] ; move16() ;
            c2 = round(L_msu(L_mult(c1,c1),s1,s1)) ; 
            s2 = round(L_mac(L_mult(s1,c1),c1,s1)) ; 
            c3 = round(L_msu(L_mult(c1,c2),s1,s2)) ; 
            s3 = round(L_mac(L_mult(s1,c2),c1,s2)) ;
            
            k = sub(ji,jr) ; 
            for( j = jr ; j < add(jr,intc) ; j++) {
                X_INT32   T0,T1,T2,T3,T4,T5,T6,T7 ;
                X_INT32   TR1,TR2,TR3 ;
                X_INT32   TI1,TI2,TI3 ;
                {
                    X_INT16 l_j, h_j, l_k, h_k ;
                    l_j = GetL15(b1[j]); 
                    h_j = GetH15(b1[j],l_j); 
                    l_k = GetL15(b1[j+k]);  
                    h_k = GetH15(b1[j+k],l_k); 
#if defined(FFT_SHL_1)
                    TR1 = Sub_Mult16x16_16(c1,l_j,h_j,s1,l_k,h_k);
                    TI1 = Add_Mult16x16_16(s1,l_j,h_j,c1,l_k,h_k);		
#else
                    TR1 = L_sub(Mult16x16_16(c1,l_j,h_j) , Mult16x16_16(s1,l_k,h_k));
                    TI1 = L_add(Mult16x16_16(s1,l_j,h_j) , Mult16x16_16(c1,l_k,h_k));
#endif
                    l_j = GetL15(b2[j]); 
                    h_j = GetH15(b2[j],l_j); 
                    l_k = GetL15(b2[j+k]); 
                    h_k = GetH15(b2[j+k],l_k);
#if defined(FFT_SHL_1)
                    TR2 = Sub_Mult16x16_16(c2,l_j,h_j,s2,l_k,h_k);
                    TI2 = Add_Mult16x16_16(s2,l_j,h_j,c2,l_k,h_k);

#else
                    TR2 = L_sub(Mult16x16_16(c2,l_j,h_j) , Mult16x16_16(s2,l_k,h_k));
                    TI2 = L_add(Mult16x16_16(s2,l_j,h_j) , Mult16x16_16(c2,l_k,h_k));
#endif
                    l_j = GetL15(b3[j]);
                    h_j = GetH15(b3[j],l_j);
                    l_k = GetL15(b3[j+k]);
                    h_k = GetH15(b3[j+k],l_k);
#if defined(FFT_SHL_1)
                    TR3 = Sub_Mult16x16_16(c3,l_j,h_j,s3,l_k,h_k);
                    TI3 = Add_Mult16x16_16(s3,l_j,h_j,c3,l_k,h_k);
#else
                    TR3 = L_sub(Mult16x16_16(c3,l_j,h_j) , Mult16x16_16(s3,l_k,h_k));
                    TI3 = L_add(Mult16x16_16(s3,l_j,h_j) , Mult16x16_16(c3,l_k,h_k));
#endif
                }

                T0 = L_add(b0[j] ,   TR2) ;
                T4 = L_add(b0[j+k] , TI2) ;
                T2 = L_sub(b0[j] ,   TR2) ;
                T6 = L_sub(b0[j+k] , TI2) ;

                T1 = L_add(TR1,TR3);
                T5 = L_add(TI1,TI3);
                T3 = L_sub(TR1,TR3);
                T7 = L_sub(TI1,TI3);

                b0[j] =   L_shr(L_add(T0 , T1),1) ; move32() ;
                b3[j+k] = L_shr(L_add(T4 , T5),1) ; move32() ;
                b2[j+k] = L_shr(L_sub(T0 , T1),1) ; move32() ;
                b1[j] =   L_shr(L_sub(T5 , T4),1) ; move32() ;
                b2[j] =   L_shr(L_sub(T2 , T7),1) ; move32() ;
                b1[j+k] = L_shr(L_add(T6 , T3),1) ; move32() ;
                b0[j+k] = L_shr(L_add(T2 , T7),1) ; move32() ;
                b3[j] =   L_shr(L_sub(T3 , T6),1) ; move32() ;
                
            }
            jthet = add(jthet,1) ;
            jr = add(jr,shl(intc,3)) ;
            ji = sub(ji,shl(intc,3)) ;
            test();
            if ( sub(ji,jl)<0 ) {
                ji = shl(sub(jr,shl(intc,1)),1);
                jl = jr ; move16();
            }
        }
    }

#endif
    Permut(b0,(X_INT16)M,N);
}


/*********************** local functions ******************************************************/

#define Perm_D(i)    tmp = L_mac(b0[i],0,0)
#define Perm(i,j)    b0[i] = b0[j] ; move32() 
#define Perm_F(i)    b0[i] = tmp ; move32()




/*----------------------------------------------------------------------------
 * FUNCTION NAME: Permut
 *
 * PURPOSE: Permutation of imaginary parts, and implements bit reverse
 *              
 *
 * INPUT:
 *      - b0 : input butterfly result
 *      - N : FFT length
 *      - M : Logarithm in base 2 of FFT length  (N = 2**M)
 *
 * OUTPUT:
 *      - b0  : FFT output  
 *
 * RETURN VALUE: none
 *   
 *
 *---------------------------------------------------------------------------*/
static void Permut(X_INT32 *b0,X_INT16 M, X_INT16 N) {
    X_INT16   jr,k ,ij,ji;
    test();
    if(sub(M,8)==0) {
    /* Tabulated permutations for N = 2**8 */
        X_INT32 tmp ;
        Perm_D(2) ;Perm(2,128) ; Perm_F(128) ;
        Perm_D(3) ;Perm(3,255) ; Perm(255,129) ; Perm_F(129) ;
        Perm_D(4) ;Perm(4,64) ; Perm_F(64) ;
        Perm_D(5) ;Perm(5,127) ; Perm(127,131) ; Perm(131,253) ; Perm(253,65) ; Perm(65,7) ;
        Perm(7,191) ; Perm(191,133) ; Perm(133,125) ; Perm(125,67) ; Perm(67,251) ; Perm(251,193) ; Perm_F(193) ;
        Perm_D(6) ;Perm(6,192) ; Perm_F(192) ;
        Perm_D(8) ;Perm(8,32) ; Perm_F(32) ;
        Perm_D(9) ;Perm(9,63) ; Perm(63,135) ; Perm(135,189) ; Perm(189,69) ; Perm(69,123) ;
        Perm(123,195) ; Perm(195,249) ; Perm(249,33) ; Perm(33,15) ; Perm(15,159) ; 
        Perm(159,141) ; Perm(141,93) ; Perm(93,75) ; Perm(75,219) ; Perm(219,201) ;
        Perm(201,57) ; Perm(57,39) ; Perm(39,183) ; Perm(183,165) ; Perm(165,117) ;
        Perm(117,99) ; Perm(99,243) ; Perm(243,225) ; Perm_F(225) ;
        Perm_D(10) ;Perm(10,160) ; Perm_F(160) ;
        Perm_D(11) ;Perm(11,223) ; Perm(223,137) ; Perm(137,61) ; Perm(61,71) ; Perm(71,187) ;
        Perm(187,197) ; Perm(197,121) ; Perm(121,35) ; Perm(35,247) ; Perm(247,161) ;
        Perm(161,13) ; Perm(13,95) ; Perm(95,139) ; Perm(139,221) ; Perm(221,73) ; 
        Perm(73,59) ; Perm(59,199) ; Perm(199,185) ; Perm(185,37) ; Perm(37,119) ; 
        Perm(119,163) ; Perm(163,245) ; Perm(245,97) ; Perm_F(97) ;
        Perm_D(12) ;Perm(12,96) ; Perm_F(96) ;
        Perm_D(14) ;Perm(14,224) ; Perm_F(224) ;
        Perm_D(17) ;Perm(17,31) ; Perm(31,143) ; Perm(143,157) ; Perm(157,77) ; Perm(77,91) ;
        Perm(91,203) ; Perm(203,217) ; Perm(217,41) ; Perm(41,55) ; Perm(55,167) ; 
        Perm(167,181) ; Perm(181,101) ; Perm(101,115) ; Perm(115,227) ; Perm(227,241) ;
        Perm_F(241) ;
        Perm_D(18) ;Perm(18,144) ; Perm_F(144) ;
        Perm_D(19) ;Perm(19,239) ; Perm(239,145) ; Perm(145,29) ; Perm(29,79) ; Perm(79,155) ;
        Perm(155,205) ; Perm(205,89) ; Perm(89,43) ; Perm(43,215) ; Perm(215,169) ; 
        Perm(169,53) ; Perm(53,103) ; Perm(103,179) ; Perm(179,229) ; Perm(229,113) ;
        Perm_F(113) ;
        Perm_D(20) ;Perm(20,80) ; Perm_F(80) ;
        Perm_D(21) ;Perm(21,111) ; Perm(111,147) ; Perm(147,237) ; Perm(237,81) ; Perm(81,27) ; 
        Perm(27,207) ; Perm(207,153) ; Perm(153,45) ; Perm(45,87) ; Perm(87,171) ; 
        Perm(171,213) ; Perm(213,105) ; Perm(105,51) ; Perm(51,231) ; Perm(231,177) ;
        Perm_F(177) ;
        Perm_D(22) ;Perm(22,208) ; Perm_F(208) ;
        Perm_D(23) ;Perm(23,175) ; Perm(175,149) ; Perm(149,109) ; Perm(109,83) ; 
        Perm(83,235) ; Perm(235,209) ; Perm(209,25) ; Perm(25,47) ; Perm(47,151) ;
        Perm(151,173) ; Perm(173,85) ; Perm(85,107) ; Perm(107,211) ; Perm(211,233) ;
        Perm(233,49) ; Perm_F(49) ;
        Perm_D(24) ;Perm(24,48) ; Perm_F(48) ;
        Perm_D(26) ;Perm(26,176) ; Perm_F(176) ;
        Perm_D(28) ;Perm(28,112) ; Perm_F(112) ;
        Perm_D(30) ;Perm(30,240) ; Perm_F(240) ;
        Perm_D(34) ;Perm(34,136) ; Perm_F(136) ;
        Perm_D(36) ;Perm(36,72) ; Perm_F(72) ;
        Perm_D(38) ;Perm(38,200) ; Perm_F(200) ;
        Perm_D(42) ;Perm(42,168) ; Perm_F(168) ;
        Perm_D(44) ;Perm(44,104) ; Perm_F(104) ;
        Perm_D(46) ;Perm(46,232) ; Perm_F(232) ;
        Perm_D(50) ;Perm(50,152) ; Perm_F(152) ;
        Perm_D(52) ;Perm(52,88) ; Perm_F(88) ;
        Perm_D(54) ;Perm(54,216) ; Perm_F(216) ;
        Perm_D(58) ;Perm(58,184) ; Perm_F(184) ;
        Perm_D(60) ;Perm(60,120) ; Perm_F(120) ;
        Perm_D(62) ;Perm(62,248) ; Perm_F(248) ;
        Perm_D(66) ;Perm(66,132) ; Perm_F(132) ;
        Perm_D(70) ;Perm(70,196) ; Perm_F(196) ;
        Perm_D(74) ;Perm(74,164) ; Perm_F(164) ;
        Perm_D(76) ;Perm(76,100) ; Perm_F(100) ;
        Perm_D(78) ;Perm(78,228) ; Perm_F(228) ;
        Perm_D(82) ;Perm(82,148) ; Perm_F(148) ;
        Perm_D(86) ;Perm(86,212) ; Perm_F(212) ;
        Perm_D(90) ;Perm(90,180) ; Perm_F(180) ;
        Perm_D(92) ;Perm(92,116) ; Perm_F(116) ;
        Perm_D(94) ;Perm(94,244) ; Perm_F(244) ;
        Perm_D(98) ;Perm(98,140) ; Perm_F(140) ;
        Perm_D(102) ;Perm(102,204) ; Perm_F(204) ;
        Perm_D(106) ;Perm(106,172) ; Perm_F(172) ;
        Perm_D(110) ;Perm(110,236) ; Perm_F(236) ;
        Perm_D(114) ;Perm(114,156) ; Perm_F(156) ;
        Perm_D(118) ;Perm(118,220) ; Perm_F(220) ;
        Perm_D(122) ;Perm(122,188) ; Perm_F(188) ;
        Perm_D(126) ;Perm(126,252) ; Perm_F(252) ;
        Perm_D(134) ;Perm(134,194) ; Perm_F(194) ;
        Perm_D(138) ;Perm(138,162) ; Perm_F(162) ;
        Perm_D(142) ;Perm(142,226) ; Perm_F(226) ;
        Perm_D(150) ;Perm(150,210) ; Perm_F(210) ;
        Perm_D(154) ;Perm(154,178) ; Perm_F(178) ;
        Perm_D(158) ;Perm(158,242) ; Perm_F(242) ;
        Perm_D(166) ;Perm(166,202) ; Perm_F(202) ;
        Perm_D(174) ;Perm(174,234) ; Perm_F(234) ;
        Perm_D(182) ;Perm(182,218) ; Perm_F(218) ;
        Perm_D(190) ;Perm(190,250) ; Perm_F(250) ;
        Perm_D(206) ;Perm(206,230) ; Perm_F(230) ;
        Perm_D(222) ;Perm(222,246) ; Perm_F(246) ;
    } else {
        /* Imaginary parts */
        jr = 2  ; move16();
        for(k = 2 ; k < M ; k++) {
            jr =shl(jr,1);
            ij = add(jr,1);
            ji = sub(shl(jr,1),1) ;
            test();
            while ( sub(ji,ij) > 0 ) {
                X_INT32 T = L_mac(b0[ij],0,0) ; 
                b0[ij] = b0[ji] ; move32() ;
                b0[ji] = T ; move32(); 
                ij = add(ij,2);
                ji = sub(ji,2);
                test() ;
            }
        }
        /*   Bit reverse */
        ji = 0 ; move16();
        ij = 0 ; move16();
        test();
        while ( sub(ij,N) < 0 ) {
            test();
            if ( sub(ij,ji) < 0 ) {
                X_INT32 T = L_mac(b0[ij],0,0) ;
                b0[ij] = b0[ji] ;  move32() ;
                b0[ji] = T ; move32();
                T = L_mac(b0[ij+1],0,0) ;
                b0[ij+1] = b0[ji+1] ; move32() ;
                b0[ji+1] = T ; move32() ;
            }
            jr = shr(N,1);
            test();
            while ( sub(jr ,ji) <= 0 ) {
                ji = sub(ji,jr); 
                jr = shr(jr,1); 
                test();
            }
            ji = add(ji,jr); 
            ij = add(ij,2) ; 
            test();
       }
    } /* M != 8 */
    return ;
 }
 

 
 
#if !defined(FFT_SHL_1)

/*----------------------------------------------------------------------------
 * FUNCTION NAME: GetL15
 *
 * PURPOSE: Return the 16 LSB of b32
 *              
 *
 * INPUT:
 *      - b32 : 32 bit input
 *
 * OUTPUT:
 *
 * RETURN VALUE: 16 LSB of b32
 *   
 *
 *---------------------------------------------------------------------------*/
static X_INT16 GetL15(X_INT32 b32) {
    return extract_l(b32); 
}
/*----------------------------------------------------------------------------
 * FUNCTION NAME: GetH15
 *
 * PURPOSE: Return the 16 MSB of b32 minus sign of bl16
 *              
 *
 * INPUT:
 *      - b32 : 32 bit input
 *      - bl16  16 bit input
 *
 * OUTPUT:
 *
 * RETURN VALUE: 16 MSB of b32 minus sign of bl16
 *   
 *
 *---------------------------------------------------------------------------*/
static X_INT16 GetH15(X_INT32 b32,X_INT16 bl16){
    return sub(extract_h(b32),shr(bl16,15));
}

/*----------------------------------------------------------------------------
 * FUNCTION NAME: Mult16x16_16
 *
 * PURPOSE: Computes 2 * a16 * bh16 +  round (a16 * bl16)
 *              
 *
 * INPUT:
 *      - a16 bh16 and bl16
 *
 * OUTPUT:
 *
 * RETURN VALUE: 2 * a16 * bh16 + round (a16 * bl16)
 *   
 *
 *---------------------------------------------------------------------------*/
static X_INT32 Mult16x16_16(X_INT16 a16,X_INT16 bl16,X_INT16 bh16) {
    return L_add(L_mult(a16,bh16),L_deposit_l(mult_r(a16,bl16))) ;

}

/*----------------------------------------------------------------------------
 * FUNCTION NAME: Mult16x16_16
 *
 * PURPOSE: Computes 2 * a16 * bh16 / 2**16
 *              
 *
 * INPUT:
 *      - a16 
 *      - b32 
 *
 * OUTPUT:
 *
 * RETURN VALUE: 2 * a16 * bh16 / 2**16
 *   
 *
 *---------------------------------------------------------------------------*/
static X_INT32 Mult16x32(X_INT16 a16,X_INT32 b32) {
    X_INT16 high ;
    X_INT16 low = extract_l(b32); 
    high  = extract_h(b32) ;
    high = sub(high,shr(low,15)) ;
    return L_add(L_mult(a16,high),L_deposit_l(mult_r(a16,low))) ;
}
#else /* FFT_SHL_1 */



/*----------------------------------------------------------------------------
 * FUNCTION NAME: GetL15
 *
 * PURPOSE: Return the 16 LSB of b32 shifted right one bit without sign extension
 *              
 *
 * INPUT:
 *      - b32 : 32 bit inputs
 *
 * OUTPUT:
 *
 * RETURN VALUE: 16 LSB of b32 shifted right one bit without sign extension
 *   
 *
 *---------------------------------------------------------------------------*/
static X_INT16 GetL15(X_INT32 b32) {
    logic16();
    return shr(extract_l(b32),1) & 0x7fff; 
}

/*----------------------------------------------------------------------------
 * FUNCTION NAME: GetH15
 *
 * PURPOSE: Return the 16 MSB of b32
 *              
 *
 * INPUT:
 *      - b32 : 32 bit input
 *
 * OUTPUT:
 *
 * RETURN VALUE: 16 MSB of b32
 *   
 *
 *---------------------------------------------------------------------------*/
static X_INT16 GetH15(X_INT32 b32,X_INT16 bl15){
    return extract_h(b32);
}

/*----------------------------------------------------------------------------
 * FUNCTION NAME: Mult16x16_16
 *
 * PURPOSE: Computes 2 * a16 * bh16 + 2 * round (a16 * bl15)
 *              
 *
 * INPUT:
 *      - a16 bh16 and bl15
 *
 * OUTPUT:
 *
 * RETURN VALUE: 2 * a16 * bh16 + 2 * round (a16 * bl15)
 *   
 *
 *---------------------------------------------------------------------------*/
static X_INT32 Mult16x16_16(X_INT16 a16,X_INT16 bl15,X_INT16 bh16) {
    return L_mac(L_mult(a16,bh16),mult_r(a16,bl15),1) ;

}
/*----------------------------------------------------------------------------
 * FUNCTION NAME: Add_Mult16x16_16
 *
 * PURPOSE: Computes 2 * xa16 * xbh16 + 2 * round (xa16 * xbl15) + 2 * ya16 * ybh16 + 2 * round (ya16 * ybl15)
 *              
 *
 * INPUT:
 *      - xa16 xbh16 and xbl15
 *      - ya16 ybh16 and ybl15
 *
 * OUTPUT:
 *
 * RETURN VALUE: 2 * xa16 * xbh16 + 2 * round (xa16 * xbl15) + 2 * ya16 * ybh16 + 2 * round (ya16 * ybl15)
 *   
 *
 *---------------------------------------------------------------------------*/

static X_INT32 Add_Mult16x16_16(X_INT16 xa16,X_INT16 xbl15,X_INT16 xbh16, X_INT16 ya16,X_INT16 ybl15,X_INT16 ybh16) {
	X_INT32 tmp = L_mac(L_mult(xa16,xbh16),mult_r(xa16,xbl15),1) ;
    return L_mac(L_mac(tmp,ya16,ybh16),mult_r(ya16,ybl15),1) ;
}
/*----------------------------------------------------------------------------
 * FUNCTION NAME: Sub_Mult16x16_16
 *
 * PURPOSE: Computes 2 * xa16 * xbh16 + 2 * round (xa16 * xbl15) - (2 * ya16 * ybh16 + 2 * round (ya16 * ybl15))
 *              
 *
 * INPUT:
 *      - xa16 xbh16 and xbl15
 *      - ya16 ybh16 and ybl15
 *
 * OUTPUT:
 *
 * RETURN VALUE: 2 * xa16 * xbh16 + 2 * round (xa16 * xbl15) - (2 * ya16 * ybh16 + 2 * round (ya16 * ybl15))
 *   
 *
 *---------------------------------------------------------------------------*/
static X_INT32 Sub_Mult16x16_16(X_INT16 xa16,X_INT16 xbl15,X_INT16 xbh16, X_INT16 ya16,X_INT16 ybl15,X_INT16 ybh16) {
	X_INT32 tmp = L_mac(L_mult(xa16,xbh16),mult_r(xa16,xbl15),1) ;
    return L_msu(L_msu(tmp,ya16,ybh16),mult_r(ya16,ybl15),1) ;
}

/*----------------------------------------------------------------------------
 * FUNCTION NAME: Mult16x32
 *
 * PURPOSE: Computes 2 * a16 * b32 / 2**16
 *              
 *
 * INPUT:
 *      - a16 
 *      - b32 
 *
 * OUTPUT:
 *
 * RETURN VALUE: 2 * a16 * b32 / 2**16
 *   
 *
 *---------------------------------------------------------------------------*/
static X_INT32 Mult16x32(X_INT16 a16,X_INT32 b32) {
    X_INT16 high ;
    X_INT16 low = shr(extract_l(b32),1) & 0x7fff ; logic16();
    high  = extract_h(b32) ;
    return L_mac(L_mult(a16,high),mult_r(a16,low),1) ;
}

#endif /* FFT_SHL_1 */

