/*=======================================================================================================*/
/*      Distributed Speech Recognition                                                                   */
/*      ETSI ES 202 050 Advanced Front-End Feature Extraction Algorithm & Compression Algorithm &        */
/*      ETSI ES 202 212 Extended Advanced Front-End Feature Extraction Algorithm & Compression Algorithm */  
/*      Bit exact C-language software implementation using ETSI basic operations library                 */
/*      3GPP TS 26.243 v 6.1.0 November 2004                                                                  */
/*      ANSI C code for the Fixed-Point Distributed Speech Recognition Extended Advanced Front-end       */
/*=======================================================================================================*/
/*---------------------------------------------------------------------------
*
* FILE NAME: ExtAdvFrontEnd.c
* PURPOSE:   This file contains the main part of the Extended Advanced DSR
*            front-end.  Speech samples are read from input waveform file
*            frame by frame.  Feature extraction is performed for each frame
*            by calling the function DoAdvProcess() (see ParmInterface.c).
*            Feature vectors are output to file in HTK format. VAD information
*            is output to file in ASCII format.  Pitch and Class information
*            are also output to files.  Command line arguments are handled by
*            a command line parsing function.
*
*---------------------------------------------------------------------------*/

/*-----------------
* File Inclusions
*-----------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "typedef.h"
#include "basic_op.h"
#include "count.h"

#if defined(WMOPS)
float FrameRateFactor=0.0001 ;
#endif
#include "fileio.h"
#include "ParmInterface_B.h"
#include "rvc_pitch_api_B.h"

/*-------------------------------
* Global Definitions and Macros
*-------------------------------*/
#define PRINTMOD 15
#define IEEE_LE 0
#define IEEE_BE 1

#define VERSION_STR "Version 6.1.0"
/*---------------
* HTK constants 
*---------------*/
#define HTK_PR_MFCC 6
#define HTK_PR_E    0x40
#define HTK_PR_0    0x2000

/*----------------------------------------------
* Global Variable Definitions and Declarations
*----------------------------------------------*/
BOOLEAN QuietMode = FALSE,      /* Supress output to stderr */
FsSpecified = FALSE,          /* Sampling frequency specified */
SwapSpecified = FALSE,        /* Byte swap for raw data files specified */
InputKindSpecified = FALSE,   /* Input file format specified */
NoOutHeaderSpecified = FALSE, /* No output HTK header option specified */
Noc0 = FALSE,                 /* No c0 coefficient to output feature vector */
NologE = FALSE;               /* No logE component to output feature vector */ 
FILE *fp_in = NULL,             /* Input HTK, NIST or raw data file */
*fp_out_B = NULL,               /* Output HTK file */
*vad_out_B = NULL;              /* Output VAD file */
FILE
*fp_out_F = NULL,               /* Output HTK file */
*vad_out_F = NULL;              /* Output VAD file */

#if defined(XAFE)
FILE *fp_out_pitch;     /* Output pitch file */
FILE *fp_out_class;     /* Output class file */
#endif
int HeaderBytesToSkip = 0;

char InFilename[199],           /* Name of input file */
InputKind[10] = "NIST";       /* Input file format */

char *ptP;
char OutFilename[199];
char VADFilename[199];

int SamplingFrequency = 8000,   /* SamplingFrequency */
NativeByteOrder = IEEE_LE,    /* Native byte ordering */
InputByteOrder,               /* Default input byte ordering */
OutParmKind=HTK_PR_MFCC + HTK_PR_E + HTK_PR_0; /* Output parameter kind MFCC_0_E as default */

int VQId;

/*------------
* Prototypes
*------------*/
static int ReadBufWave (FILE *fp_in, FILE_TYPE *buf, int nSamples, int Swap);
static BOOLEAN ParseCommLine (int argc, char *argv[]);

/*-----------
* Functions
*-----------*/
/*----------------------------------------------------------------------------
* FUNCTION NAME: ReadBufWave
*
* PURPOSE:       Read buffer from input file
*
* INPUT:
*   fp_in        Pointer to file
*   nSamples     Number of samples to read
*   swap         Byte swap specification
*
* OUTPUT
*   buf          Pointer to buffer
*
* RETURN VALUE
*   FALSE        In case of fread() error
*   TRUE         Otherwise
*
*---------------------------------------------------------------------------*/
static int ReadBufWave (FILE *fp_in, FILE_TYPE *buf, int nSamples, int Swap)
{
    FILE_TYPE s;
    int i;
    
    for (i=0; i<nSamples; i++)
    {
        if (fread (&s, sizeof (s), 1, fp_in) != 1) return FALSE;
        if (Swap) s = ((s & 0x00ff) << 8) | ((s & 0xff00) >> 8);
        buf[i] = (FILE_TYPE)s;
    }
    return TRUE;
}

/*----------------------------------------------------------------------------
* FUNCTION NAME: ParseCommLine
*
* PURPOSE:       Parses command line arguments, opens input and output files
*
* INPUT:
*   argc         Number of command line arguments
*   argv         Array of command line arguments
*
* OUTPUT
*   none
*
* RETURN VALUE
*   FALSE        In case of any errors
*   TRUE         Otherwise
*
*---------------------------------------------------------------------------*/
static BOOLEAN ParseCommLine (int argc, char *argv[])
{
    int mark = 0, mand_arg_no = 1;
    
    for ( ; argc; argc--, mark++)
    {
        if (strcmp (argv[mark], "-q") == 0)
        {
            QuietMode = TRUE;
        }
        else if (strcmp (argv[mark], "-fs") == 0)
        {
            FsSpecified = TRUE;
            --argc;
            ++mark;
            SamplingFrequency = 1000 * atoi (argv[mark]);
        }
        else if (strcmp (argv[mark], "-swap") == 0)
        {
            SwapSpecified = TRUE;
        }
        else if (strcmp (argv[mark], "-F") == 0)
        {
            InputKindSpecified = TRUE;
            --argc;
            ++mark;
            strcpy (InputKind, argv[mark]);
        }
        else if (strcmp (argv[mark], "-noh") == 0)
        {
            NoOutHeaderSpecified = TRUE;
        }
        else if (strcmp (argv[mark], "-noc0") == 0)
        {
            Noc0 = TRUE;
        }
        else if (strcmp (argv[mark], "-nologE") == 0)
        {
            NologE = TRUE;
        }
        else if (strcmp (argv[mark],"-skip_header_bytes") == 0)
        {
          --argc;
          ++mark;
          HeaderBytesToSkip = atoi (argv[mark]);
        }
        else if (argv[mark][0] == '-')
        {
            fprintf (stderr, "WARNING:  Un-recognized flag '%s' !\r\n",
                     argv[mark]);
        }
        else
        {
            switch (mand_arg_no)
            {

            case 1: /* First mandatory argument - input file */
                strcpy (InFilename, argv[mark]);
                fp_in = fopen (argv[mark], "rb");
                if (fp_in == NULL)
                {
                    fprintf (stderr, "ERROR:  Could not open file '%s' !\r\n",
                             argv[mark]);
                    return FALSE;
                }
                break;

            case 2: /* Second mandatory argument - output HTK file */
                strcpy (OutFilename, argv[mark]);
                strcpy(VADFilename,argv[mark]);
                if( (ptP= strrchr(VADFilename,'.')) != NULL) {
                    *ptP = '\0' ;
                } 
                strcat(VADFilename,".vad");
                fp_out_B = fopen (OutFilename, "wb");
                if (fp_out_B == NULL)
                {
                    fprintf (stderr, "ERROR: Could not open file '%s' !\r\n",
                             OutFilename);
                    return FALSE;
                }
                vad_out_B = fopen (VADFilename, "wb");
                if (vad_out_B == NULL)
                {
                    fprintf (stderr, "ERROR: Could not open file '%s' !\r\n",
                             VADFilename);
                    return FALSE;
                }       
                strcpy (OutFilename, argv[mark]);
                if( (ptP= strrchr(OutFilename,'.')) != NULL) {
                    memmove(ptP+2,ptP,strlen(ptP)+1);
                    *ptP++= '_' ;
                    *ptP = 'F' ;

                } else {
                    strcat(OutFilename,"_F");
                }
                strcpy(VADFilename,OutFilename);
                if( (ptP= strrchr(VADFilename,'.')) != NULL) {
                    *ptP = '\0' ;
                } 
                strcat(VADFilename,".vad");
                break;
#if defined(XAFE)
            case 3: /* Third mandatory argument - output pitch file */
                strcpy (OutFilename, argv[mark]);
                fp_out_pitch = fopen (OutFilename,"wb");
                if (fp_out_pitch == NULL)
                {
                  fprintf (stderr, "ERROR: Could not open file '%s' !\r\n",
                           OutFilename);
                  return FALSE;
                }
                break;

            case 4: /* Fourth mandatory argument - output class file */
                strcpy (OutFilename, argv[mark]);
                fp_out_class = fopen (OutFilename,"wb");
                if (fp_out_class == NULL)
                {
                  fprintf (stderr, "ERROR: Could not open file '%s' !\r\n",
                           OutFilename);
                  return FALSE;
                }
                break;
#endif                
            default:
                fprintf (stderr, "ERROR: Too many arguments !\r\n");
                return FALSE;

            } /* end of switch */
            mand_arg_no++;

        }

    }
#if defined(XAFE)   
    if (!fp_in || !fp_out_B || !fp_out_pitch || !fp_out_class)
    {
        fprintf (stderr,"ERROR: Input, output, pitch, and class files must be given!\r\n");
        return FALSE;
    }
#else
    if (!fp_in || !fp_out_B )
    {
        fprintf (stderr,"ERROR: Input, output files must be given!\r\n");
        return FALSE;
    }
#endif
    if (strcmp (InputKind, "NIST") && strcmp (InputKind, "HTK") && strcmp (InputKind, "RAW"))
    {
        fprintf (stderr, "ERROR:   Invalid input file format '%s'!\r\n", InputKind);
        return FALSE;
    }

    if (strcmp (InputKind, "RAW") && HeaderBytesToSkip != 0)
    {
        fprintf (stderr, "ERROR:  skip_header_bytes can be specified for RAW format only. \r\n");
        return FALSE;
    }
    
    if (strcmp (InputKind, "RAW") && FsSpecified)
        fprintf (stderr, "WARNING:   Sampling frequency needs to be specified only for raw data files.\r\n");
    
    if (strcmp (InputKind, "RAW") && SwapSpecified)
        fprintf (stderr, "WARNING:   Byte swapping needs to be specified only for raw data files if necessary.\r\n");
    
    if (Noc0 && NologE) TextToParmKind( "MFCC", &OutParmKind);
    else if (Noc0) TextToParmKind( "MFCC_E", &OutParmKind);
    else if (NologE) TextToParmKind( "MFCC_0", &OutParmKind);
    
    return TRUE;
}

/*----------------------------------------------------------------------------
* FUNCTION NAME: main
*
* PURPOSE:       Main front-end operations from input speech samples to output
*                feature vectors. See embedded comments.
*
* INPUT:
*   argc         Number of command line arguments (passed to ParseCommLine)
*   argv         Array of command line arguments (passed to ParseCommLine)
*
* OUTPUT
*   none
*
* RETURN VALUE
*   TRUE         In case of any errors
*   FALSE        Otherwise
*
*---------------------------------------------------------------------------*/
int main (int argc, char *argv[])
{
    FILE_TYPE *SigBuf;
    COEF_TYPE FeatureBuffer_B[NUM_CEP_COEFF + 3]; /* +2 as required by Vad */
    FEParamsX_F  *pFEParX_B;
    HTK_Header InHheader, OutHheader;
    NIST_Header InNheader;
    
    int NbSamplesToRead;

#if defined(XAFE)
    int rc;
#endif
    
    long FrameCounter = 0;
    long SpeechFrameCounter = 0;
    long NonSpeechFrameCounter = 0;

#if defined(WMOPS)
    VQId=getCounterId("AdvFrontEnd");
    setCounter(VQId);
    Init_WMOPS_counter();
#endif
    fprintf (stderr,"\r\nExtended Advanced DSR Front-End %s\r\n", VERSION_STR);
    
    if (!ParseCommLine (argc - 1, argv + 1))
    {
#if defined(XAFE)
      fprintf (stderr,"\r\n    ETSI DSR Extended Advanced Front-End Feature Extraction Algorithm");
#else
      fprintf (stderr,"\r\n    ETSI DSR Advanced Front-End Feature Extraction Algorithm");
#endif
      fprintf (stderr,"\r\n    Bit-Exact Implementation");
      fprintf (stderr,"\r\n    %s\r\n\n",VERSION_STR);

      fprintf (stderr, "\r\n    USAGE:");
      fprintf (stderr, "  %s infile HTK_outfile"
#if defined(XAFE)
        " pitch_outfile class_outfile" 
#endif
        " [options]\r\n", argv[0]);
      fprintf (stderr, "\r\n    OPTIONS:\r\n");
      fprintf (stderr, "     -q            Quiet Mode (%s)\r\n", QuietMode ? "TRUE" : "FALSE");
      fprintf (stderr, "     -F    format  Input file format (NIST,HTK,RAW) (%s)\r\n", InputKind);
      fprintf (stderr, "     -fs   freq    Sampling frequency in kHz (%d,%d,%d) (%d)\r\n",SAMPLING_FREQ_1,SAMPLING_FREQ_2,SAMPLING_FREQ_3,SamplingFrequency/1000);
      fprintf (stderr, "     -swap         Change input byte ordering (%s)\r\n", SwapSpecified ? "Swapped" : "Native");
      fprintf (stderr, "                   (Native byte ordering is %s)\r\n",NativeByteOrder ? "ieee-be" : "ieee-le");
      fprintf (stderr, "     -noh          No HTK header to output file (%s)\r\n", NoOutHeaderSpecified ? "TRUE" : "FALSE" );
      fprintf (stderr, "     -noc0         No c0 coefficient to output feature v ector (%s)\r\n", Noc0 ? "TRUE" : "FALSE" );
      fprintf (stderr, "     -nologE       No logE component to output feature v ector (%s)\r\n", NologE ? "TRUE" : "FALSE" );
      fprintf (stderr, "     -skip_header_bytes n - Skip header, first n bytes ( Only for F RAW)\r\n");

      return TRUE;
    }

    if (!QuietMode)
    {
      fprintf (stderr,"\r\n    ETSI DSR Extended Advanced Front-End Feature Extraction Algorithm");
      fprintf (stderr,"\r\n    Bit-Exact Implementation");
      fprintf (stderr,"\r\n    %s\r\n\n",VERSION_STR);
    }

    /*----------------*/
    /* Initialization */
    /*----------------*/
    InputByteOrder = NativeByteOrder;
    
    /*-----------------------------------------------------------------
       * Read input header, extract sampling frequency and byte ordering
        *-----------------------------------------------------------------*/
    if (!strcmp (InputKind, "NIST"))
    {
        if (!ReadNISTHeader (fp_in, &InNheader))
        {
            fprintf (stderr, "ERROR:   Invalid NIST header !\r\n");
            goto _FaultExit;
        }
        SamplingFrequency = InNheader.SampleRate;
        if (strcmp (InNheader.SampleByteFormat, "10"))
            InputByteOrder = IEEE_LE;
        else
            InputByteOrder = IEEE_BE;
    }
    else if (!strcmp (InputKind, "HTK"))
    {
        if (!ReadHTKHeader (fp_in, &InHheader, InputByteOrder!=IEEE_BE))
        {
            fprintf (stderr, "ERROR:   Invalid HTK header !\r\n");
            goto _FaultExit;
        }
        // 625->16kHz, 1250->8kHz, 909->11kHz
        SamplingFrequency = 10 * floor ((float) 1e6 / (float) InHheader.sampPeriod);
        InputByteOrder = IEEE_BE;
    }
        
        /*------------------------------------------------------------
        * Write output header (number of frames to be updated later)
        *------------------------------------------------------------*/
        OutHheader.nSamples = 0;
        OutHheader.sampPeriod = 100000; /* 10000.0 us (100 Hz) */
        OutHheader.sampSize = (NUM_CEP_COEFF - (Noc0 ? 1:0) + (NologE ? 0:1)) * 4;
        OutHheader.sampKind = OutParmKind;      
        if (!NoOutHeaderSpecified) WriteHTKHeader (fp_out_B, &OutHheader);

        /*------------------------------------------
        * Memory allocation for FE data structures
        *------------------------------------------*/
        pFEParX_B = AdvProcessAlloc_B (SamplingFrequency);

        /*-------------------------------------------------------
        * Initialization of FE data structures and input buffer
        *-------------------------------------------------------*/
        AdvProcessInit_B (pFEParX_B);

#if defined(XAFE)
        rc = RVC_ConstructPitchRom_be(SAMPLING_FREQ_1*1000,0,
                                      &(pFEParX_B->pRvcPitchRom));
        if (rc != RVC_OK)
        {
          fprintf(stderr,"ERROR: Can't initialize RVC_PITCH_ROM,RC = %d\r\n",rc);
          goto _FaultExit;
        }
        rc = RVC_ConstructPitchMeter_be(pFEParX_B->pRvcPitchRom,
                                        &(pFEParX_B->pRvcPitchMeter));
        if ( rc!= RVC_OK)
        {
          fprintf (stderr, "ERROR: can't initialize RVC_PITCH_METER,RC = %d\r\n",rc);
          goto _FaultExit;
        }

        pFEParX_B->pitch_fid = fp_out_pitch;
        pFEParX_B->class_fid = fp_out_class;
#endif
        
        NbSamplesToRead  = pFEParX_B->NbSamplesToRead;
        
        SigBuf = calloc (1, sizeof (SigBuf[0]) * NbSamplesToRead);
        if (SigBuf == NULL)
        {
            fprintf (stderr, "ERROR:   Memory allocation error occured!\r\n");
            goto _FaultExit;
        }

        
        /*------------
        * Processing
        *------------*/

        fseek(fp_in,HeaderBytesToSkip,SEEK_SET);

        while (ReadBufWave (fp_in, SigBuf, NbSamplesToRead,
            (SwapSpecified || InputByteOrder != NativeByteOrder)))
        {
            FrameCounter++;
#if defined(WMOPS)
            fwc();
            WMOPS_output(0);
            Reset_WMOPS_counter(); 
#endif

            if (DoAdvProcess_B (SigBuf, FeatureBuffer_B, pFEParX_B))
            {
               /*---------------
                * Output result
                *---------------*/
                WriteHTKFeature (fp_out_B, FeatureBuffer_B, (short) (NUM_CEP_COEFF - (Noc0 ? 1:0) + (NologE ? 0:1)));
                
                fprintf (vad_out_B, "%d ", pFEParX_B->VAD); 
                if (pFEParX_B->VAD == SPEECH_FRAME)
                    SpeechFrameCounter++;
                else
                    NonSpeechFrameCounter++;
            }
            else
            {
/*              goto _FaultExit; */
            }
        
            /*---------------------------
            * Display processing status
            *---------------------------*/
            if (!QuietMode && !(FrameCounter % PRINTMOD))
            {
                fprintf (stderr, "\rProcessing status: %ld frames ...", FrameCounter);
                fflush (stderr);
            }   
        }
        
#if defined(WMOPS)
        fwc();
        WMOPS_output(0);
        Reset_WMOPS_counter(); 
#endif

        /*----------
        * Flushing
        *----------*/
        while (FlushAdvProcess_B (FeatureBuffer_B, pFEParX_B))
        {       
        /*---------------
         * Output result
         *---------------*/
            WriteHTKFeature (fp_out_B, FeatureBuffer_B, (short) (NUM_CEP_COEFF - (Noc0 ? 1:0) + (NologE ? 0:1)));
#if defined(WMOPS)
            fwc();
            WMOPS_output(0);
            Reset_WMOPS_counter(); 
#endif

            fprintf (vad_out_B, "%d ", pFEParX_B->VAD); 
            if (pFEParX_B->VAD == SPEECH_FRAME)
                SpeechFrameCounter++;
            else
                NonSpeechFrameCounter++;
        }
        /*----------------
        * Memory release
        *----------------*/
#if defined(XAFE)
        RVC_DestructPitchRom_be(pFEParX_B->pRvcPitchRom);
        RVC_DestructPitchMeter_be(pFEParX_B->pRvcPitchMeter);
#endif
        AdvProcessDelete_B (&pFEParX_B);
        free (SigBuf);

#if defined(WMOPS)
        WMOPS_output(1);
#endif
        /*----------------------------------------
        * Correct number of frames in HTK header
        *----------------------------------------*/
        OutHheader.nSamples = SpeechFrameCounter + NonSpeechFrameCounter;
        if (!NoOutHeaderSpecified) WriteHTKHeader (fp_out_B, &OutHheader);
        
        /*------------------------------
        * Close input and output files
        *------------------------------*/
        fclose (fp_in);
        fclose (fp_out_B);
        fclose (vad_out_B);
#if defined(XAFE)
        fclose(fp_out_pitch);
        fclose(fp_out_class);
#endif        
        /*----------------------
        * Display final status 
        *----------------------*/
        if (!QuietMode)
            fprintf (stderr, "\rProcessed: %ld Frames.           \r\n", FrameCounter);
        
        fprintf (stderr, "Extended Advanced Front End: ");
        fprintf (stderr, InFilename);
        fprintf (stderr, ": %ld Frames, %ld Speech, %ld Non Speech, %ld Lost.\r\n", FrameCounter, 
            SpeechFrameCounter, NonSpeechFrameCounter, FrameCounter - SpeechFrameCounter - NonSpeechFrameCounter);  
        
        if (SpeechFrameCounter == 0)
            fprintf (stderr, "NO SPEECH DETECTED !\r\n");
        
    
        printf("Exiting\n");
        exit (0);
    
_FaultExit:
    if (fp_in)
        fclose (fp_in);
    if (fp_out_B)
        if (fclose (fp_out_B))
            fprintf (stderr, "Can't rewrite output file!\r\n");
#if defined(XAFE)
    if (fp_out_pitch)
        fclose(fp_out_pitch);
    if (fp_out_class)
        fclose(fp_out_class);
#endif
        return TRUE;
}
