/*=======================================================================================================*/
/*      Distributed Speech Recognition                                                                   */
/*      ETSI ES 202 050 Advanced Front-End Feature Extraction Algorithm & Compression Algorithm &        */
/*      ETSI ES 202 212 Extended Advanced Front-End Feature Extraction Algorithm & Compression Algorithm */  
/*      Bit exact C-language software implementation using ETSI basic operations library                 */
/*      3GPP TS 26.243 v 6.1.0 November 2004                                                                  */
/*      ANSI C code for the Fixed-Point Distributed Speech Recognition Extended Advanced Front-end       */
/*=======================================================================================================*/


/*
**********************************************************************
*                                                                    *
*                           Include Files                            *
*                                                                    *
**********************************************************************
*/

#include <assert.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>

#include "dsrAfeVad_B.h"
#include "fix_mathlib.h"
#include "typedef.h"
#include "basic_op.h"
#include "count.h"


/*
**********************************************************************
*                                                                    *
*                              Defines                               *
*                                                                    *
**********************************************************************
*/

/*
* Number of channels
*/

#define			NUM_CHAN		23


/*
* Minimum channel energy and Initial signal energy
* (Corresponds roughly to a signal at -36 dB)
*/

#define			MIN_CH_ENRG_MANTISSA	20000
#define			MIN_CH_ENRG_SHIFT	25
#define			INIT_SIG_ENRG_MANTISSA	30518
#define			INIT_SIG_ENRG_SHIFT	8


/*
* Channel energy and channel noise energy smoothing factors
*/

#define			CE_SM_FAC		18022
#define			CE_SM_FAC_COMPL		14746
#define			CNE_SM_FAC		3277
#define			CNE_SM_FAC_COMPL	29491


/*
* Low and high values of "gamma" used in filtering
* "long term log-spectral energy"
*/

#define			LO_GAMMA		22938
#define			LO_GAMMA_COMPL		9830
#define			HI_GAMMA		29491
#define			HI_GAMMA_COMPL		3277


/*
* Low and high values of "beta" used in filtering "snr"
*/

#define			LO_BETA			31130
#define			HI_BETA			32702


/*
* Number of initial frames, which are assumed to be non-speech
*/

#define			INIT_FRAMES		10


/*
* Sine start channel and peak-to-average threshold -
* Used in "sine wave detection"
*/

#define			SINE_START_CHAN		4
#define			PEAK_TO_AVE_THLD	10


/*
* Deviation threshold, hysteresis count threshold, and forced
* update count threshold - Used in "forced update" of channel
* noise energy estimates
*/

#define			DEV_THLD		1523942
#define			HYSTER_CNT_THLD		9
#define			F_UPDATE_CNT_THLD	500


/*
* Non-speech threshold
*/

#define			NON_SPEECH_THLD		32



/*
***********************************************************************
*
*   FUNCTION NAME: dsr_afe_vad
*
*   PURPOSE: To detect Voice Activity in DSR advanced front-end
*
*   INPUTS:
*
*      pswMFBOutArray[] - Array of Mel-Filter Bank outputs
*
*      iMFBOutShift - Shift parameter for the above array
*
*   OUTPUTS:
*
*      *piHangOverFlag - This flag is set if the current frame is
*           a hang-over frame
*
*   RETURN VALUE:
*
*      The VAD value is returned
*
**********************************************************************
*/


/*
**********************************************************************
*                                                                    *
*                            The function                            *
*                                                                    *
**********************************************************************
*/

X_INT16 dsr_afe_vad(Shortword *pswMFBOutArray, X_INT16 iMFBOutShift, 
                    X_INT16 *piHangOverFlag)
{


/*
********************************************************************
*                                                                  *
*                         Static variables                         *
*                                                                  *
********************************************************************
*/

static const X_INT16 piBurstConst[20] = {2, 2, 3, 3, 4, 4, 4, 4, 5, 5,
5, 5, 5, 5, 6, 6, 6, 6, 6, 6};

static const X_INT16 piHangConst[20] = {54, 52, 50, 48, 46, 44, 42, 40, 38, 36,
34, 32, 30, 28, 26, 24, 22, 20, 18, 16};

static const X_INT16 piVADThld[20] = {32, 34, 36, 38, 40, 42, 44, 46, 48, 50,
52, 54, 55, 56, 57, 57, 58, 58, 58, 58};


static X_INT16 iBurstConst = 6;
static X_INT16 iBurstCount = 0;

static X_INT16 iHangConst = 28;
static X_INT16 iHangCount = -1;

static X_INT16 iVADThld = 56;


/*
********************************************************************
*                                                                  *
*                        Automatic variables                       *
*                                                                  *
********************************************************************
*/

X_INT16 iQSnr;
X_INT16 iVad;
X_INT16 iVoiceMetric;
X_INT16 iVvad;

/*
 * Function declaration
 */

X_INT16 get_vm(Shortword *, X_INT16, X_INT16 *);



/*
********************************************************************
*                                                                  *
*                         Executable Code                          *
*                                                                  *
********************************************************************
*/


/*
* Get the "voice metric" for the frame
*/

iVoiceMetric = get_vm(pswMFBOutArray,iMFBOutShift,&iQSnr);


/*
* Estimate "voice activity" for the frame
*/

test();
//iVvad = (iVoiceMetric > iVADThld)? TRUE : FALSE;	move16();
iVvad = (sub(iVoiceMetric, iVADThld) >0 )? TRUE : FALSE; move16();


/*
* Add Hangover
*/

test();
if (iVvad == TRUE)
{
  iBurstCount = add(iBurstCount,1);
}
else
{
  iBurstCount = 0;		move16();
}

test();
if (sub(iBurstCount, iBurstConst) >=0 )
{
  iHangCount = iHangConst;			move16();
  iBurstCount = iBurstConst;			move16();
}


/*
* Make the "vad" decision for the frame
*/

test(); test(); logic16();  /* workaround: test to FALSE instead */
if ((iVvad == TRUE) || (iHangCount >= 0))
{
  iVad = TRUE;					move16();
}
else
{
  iVad = FALSE;					move16();
}

test(); test();
if ((iVvad == FALSE) && (iHangCount >= 0))
{
  *piHangOverFlag = TRUE;			move16();
}
else
{
  *piHangOverFlag = FALSE;			move16();
}

test();
if (iHangCount >= 0)
{
  iHangCount = sub(iHangCount,1);
}


/*
* Update the thresholds and return
*/

iVADThld = piVADThld[iQSnr];			move16();
iBurstConst = piBurstConst[iQSnr];		move16();
iHangConst = piHangConst[iQSnr];		move16();

return(iVad);

}



/*
***********************************************************************
*
*   FUNCTION NAME: get_vm
*
*   PURPOSE: To estimate the "voice metric" of the frame
*
*   INPUTS:
*
*      pswMFBOutArray[0:NUM_CHAN-1] - Array of Mel-Filter bank outputs
*
*      iMFBOutShift - Shift parameter for the above array
*
*   OUTPUTS:
*
*      *piQSnr - Current estimate of the quantized SNR
*
*   RETURN VALUE:
*
*      iVoiceMetric is returned
*
**********************************************************************
*/


/*
**********************************************************************
*                                                                    *
*                            The function                            *
*                                                                    *
**********************************************************************
*/

X_INT16 get_vm(Shortword *pswMFBOutArray, X_INT16 iMFBOutShift,
               X_INT16 *piQSnr)
{


/*
********************************************************************
*                                                                  *
*                         Static variables                         *
*                                                                  *
********************************************************************
*/

/*
* The voice metric table is defined below.  It is a non-
* linear table that maps the SNR index (quantized SNR value)
* to a number that is a measure of voice quality.
*/

static const X_INT16 piVMTable[90] =
{
1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3,
4, 4, 4, 5, 5, 5, 6, 6, 7, 7, 7, 8, 8, 9, 9, 10,
10, 11, 12, 12, 13, 13, 14, 15, 15, 16, 17, 17,
18, 19, 20, 20, 21, 22, 23, 24, 24, 25, 26, 27,
28, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 37,
38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49,
50, 50, 50, 50, 50, 50, 50, 50, 50, 50
};


/*
* The different thresholds (as a function of SNR) are
* defined below
*/

static const X_INT16 piSigThld[20] = {36, 43, 52, 62, 73, 86, 101, 117, 134, 153,
173, 194, 217, 242, 268, 295, 295, 295, 295, 295};

static const X_INT16 piUpdateThld[20] = {31, 32, 33, 34, 35, 36, 37, 37, 37, 37,
37, 37, 37, 37, 37, 38, 38, 38, 38, 38};


static X_INT16 iFrameCount = 0;
static X_INT16 iFUpdateFlag = FALSE;
static X_INT16 iHysterCount = 0;
static X_INT16 iLastUpdateCount = 0;
static X_INT16 iSigThld = 217;
static X_INT16 iUpdateCount = 0;
static X_INT16 iChanEnrgShift;
static X_INT16 iChanNoiseEnrgShift;


/*
* The shape table is defined below.  It is used to correct
* the spectral shaping caused by the different channel
* widths used in DSR-AFE standard.
*/

static const Shortword pswShapeTable[NUM_CHAN] =
{
  10923,
  10923,
  9362,
  9362,
  9362,
  8192,
  8192,
  7281,
  6554,
  6554,
  6554,
  5957,
  5462,
  5040,
  4683,
  4683,
  4368,
  3854,
  3641,
  3641,
  3277,
  2979,
  2851
};

static Shortword pswChanEnrg[NUM_CHAN];
static Shortword pswChanNoiseEnrg[NUM_CHAN];
static Shortword swBeta = LO_BETA;
static Shortword swSnr;

static NormSw pnsLogSpecEnrgLong[NUM_CHAN];


/*
********************************************************************
*                                                                  *
*                        Automatic variables                       *
*                                                                  *
********************************************************************
*/

X_INT16 i,j;
X_INT16 iUpdateThld;
X_INT16 iFirstFrameFlag;
X_INT16 iInitFramesFlag;
X_INT16 iPeak2AveFlag;
X_INT16 iLogSpecEnrgDevFlag;
X_INT16 iUpdateFlag;
X_INT16 iQSnr;
X_INT16 iVoiceMetric;
X_INT16 iLog2;
X_INT16 iTemp;
X_INT16 iShift;
X_INT16 iChanSnr;
X_INT16 iNoiseEnrgShift;
X_INT16 iSigEnrgInstShift;

X_INT16 piChanSnr[NUM_CHAN];

Shortword swGamma;
Shortword swGammaC;
Shortword swTemp;
Shortword swTemp1;
Shortword swTemp2;
Shortword swMax;
Shortword swPeak;
Shortword swLog2;
Shortword swMantissa;
Shortword swNoiseEnrg;
Shortword swSigEnrgInst;
Shortword swSnrInst;
Shortword swBetaC;

Longword lwTemp;
Longword lwTemp1;
Longword lwTemp2;
Longword lwPeak;
Longword lwAve;
Longword lwMantissa;
Longword lwShift;
Longword lwLogSpecEnrgDev;
Longword lwNoiseEnrg;
Longword lwSigEnrgInst;
Longword lwSnrInst;

NormSw pnsLogSpecEnrg[NUM_CHAN];


/*
********************************************************************
*                                                                  *
*                         Executable Code                          *
*                                                                  *
********************************************************************
*/


/*
* Increment the frame counter and set flags
*/

iFrameCount = add(iFrameCount,1);

test();
if (sub(iFrameCount,0x7FFF) == 0)
{
  iFrameCount = 0x7FFE;				move16();
}

test();
iFirstFrameFlag = (sub(iFrameCount,1) == 0);
test();
iInitFramesFlag = (sub(iFrameCount,INIT_FRAMES) <= 0);


/*
* Estimate the energy in each channel
*/

test();
if (iFirstFrameFlag)
{
  for (i = 0; i < NUM_CHAN; i++)
  {
    pswChanEnrg[i] = round(L_mult(pswMFBOutArray[i],pswShapeTable[i]));
						move16();
  }
  iChanEnrgShift = iMFBOutShift;		move16();
}
else
{
  test();
  if (sub(iMFBOutShift, iChanEnrgShift) >=0 )
  {
    for (i = 0; i < NUM_CHAN; i++)
    {
      swTemp = round(L_shr(L_mult(pswMFBOutArray[i],pswShapeTable[i]),
                        sub(iMFBOutShift,iChanEnrgShift)));
      pswChanEnrg[i] = round(L_mac(L_mult(swTemp,(Shortword)
                       CE_SM_FAC),pswChanEnrg[i],(Shortword)CE_SM_FAC_COMPL));
						move16();
    }
  }
  else
  {
    for (i = 0; i < NUM_CHAN; i++)
    {
      swTemp = round(L_mult(pswMFBOutArray[i],pswShapeTable[i]));
      lwTemp = L_shr(L_mult(pswChanEnrg[i],(Shortword)CE_SM_FAC_COMPL),
                         sub(iChanEnrgShift,iMFBOutShift));
      pswChanEnrg[i] = round(L_mac(lwTemp,swTemp,(Shortword)CE_SM_FAC));
      						move16();
    }
    iChanEnrgShift = iMFBOutShift;		move16();
  }

}

swMax = pswChanEnrg[0];				move16();
for (i = 1; i < NUM_CHAN; i++)
{
  test();
  if (sub(swMax,pswChanEnrg[i]) < 0)
  {
    swMax = pswChanEnrg[i];			move16();
  }
}
iShift = norm_s(swMax);
for (i = 0; i < NUM_CHAN; i++)
{
  pswChanEnrg[i] = shl(pswChanEnrg[i],iShift);	move16();
}
iChanEnrgShift = add(iChanEnrgShift,iShift);

iTemp = sub(iChanEnrgShift,MIN_CH_ENRG_SHIFT);
test();
if (iTemp > 0)
{
  for (i = 0; i < NUM_CHAN; i++)
  {
    pswChanEnrg[i] = (Shortword)MIN_CH_ENRG_MANTISSA;	move16();
  }
  iChanEnrgShift = MIN_CH_ENRG_SHIFT;			move16();
}
else
{
  for (i = 0; i < NUM_CHAN; i++)
  {
    swTemp = shr(MIN_CH_ENRG_MANTISSA,-iTemp);
    test();
    if (sub(pswChanEnrg[i], swTemp) < 0)
    {
      pswChanEnrg[i] = swTemp;				move16();
    }
  }
}


/*
 * Calculate the Peak-to-Average ratio and set flag
 */

/*
 * Compute (NUM_CHAN*PEAK)
 */

swPeak = pswChanEnrg[SINE_START_CHAN];		move16();
for (i = SINE_START_CHAN + 1; i < NUM_CHAN; i++)
{
  test();
  if (sub(pswChanEnrg[i],swPeak) > 0)
  {
    swPeak = pswChanEnrg[i];			move16();
  }
}
lwPeak = L_mult(swPeak,(Shortword)NUM_CHAN);


/*
 * Compute NUM_CHAN*AVE*PEAK_TO_AVE_THLD
 */

lwAve = L_mult(pswChanEnrg[0],(Shortword)PEAK_TO_AVE_THLD);
for (i = 1; i < NUM_CHAN; i++)
{
  lwAve = L_mac(lwAve,pswChanEnrg[i],(Shortword)PEAK_TO_AVE_THLD);
}


/*
 * Set flag if (PEAK >= AVE*PEAK_TO_AVE_THLD), that is,
 * (NUM_CHAN*PEAK) >= (NUM_CHAN*AVE*PEAK_TO_AVE_THLD)
 */

test();
iPeak2AveFlag = (L_sub(lwPeak,lwAve) >= 0);


/*
* Estimate the channel noise energies from the first
* INIT_FRAMES frames
*/

test(); test();
if ((iInitFramesFlag == TRUE) || (iFUpdateFlag == TRUE))
{
  test();
  if (iPeak2AveFlag == FALSE)
  {
    test();
    if (iFirstFrameFlag == TRUE)
    {
      for (i = 0; i < NUM_CHAN; i++)
      {
        pswChanNoiseEnrg[i] = pswChanEnrg[i];		move16();
      }
      iChanNoiseEnrgShift = iChanEnrgShift;		move16();
    }
    else
    {
      test(); 
      if (sub(iChanEnrgShift, iChanNoiseEnrgShift) >=0 )
      {
        for (i = 0; i < NUM_CHAN; i++)
        {
          lwTemp = L_shr(L_mult(pswChanEnrg[i],(Shortword)9830),
                            sub(iChanEnrgShift,iChanNoiseEnrgShift));
          pswChanNoiseEnrg[i] = round(L_mac(lwTemp,pswChanNoiseEnrg[i],
                                        (Shortword)22938));	move16();
        }
      }
      else
      {
        for (i = 0; i < NUM_CHAN; i++)
        {
          lwTemp = L_shr(L_mult(pswChanNoiseEnrg[i],(Shortword)22938),
                            sub(iChanNoiseEnrgShift,iChanEnrgShift));
          pswChanNoiseEnrg[i] = round(L_mac(lwTemp,pswChanEnrg[i],
                                                 (Shortword)9830)); move16();
        }
        iChanNoiseEnrgShift = iChanEnrgShift;		move16();
      }

      swMax = pswChanNoiseEnrg[0];			move16();
      for (i = 1; i < NUM_CHAN; i++)
      {
	test();
        if (sub(swMax,pswChanNoiseEnrg[i]) < 0)
        {
          swMax = pswChanNoiseEnrg[i];			move16();
        }
      }
      iShift = norm_s(swMax);
      for (i = 0; i < NUM_CHAN; i++)
      {
        pswChanNoiseEnrg[i] = shl(pswChanNoiseEnrg[i],iShift);	move16();
      }
      iChanNoiseEnrgShift = add(iChanNoiseEnrgShift,iShift);

      iTemp = sub(iChanNoiseEnrgShift,MIN_CH_ENRG_SHIFT);
      test();
      if (iTemp > 0)
      {
        for (i = 0; i < NUM_CHAN; i++)
        {
          pswChanNoiseEnrg[i] = (Shortword)MIN_CH_ENRG_MANTISSA;move16();
        }
        iChanNoiseEnrgShift = MIN_CH_ENRG_SHIFT;		move16();
      }
      else
      {
        for (i = 0; i < NUM_CHAN; i++)
        {
          swTemp = shr(MIN_CH_ENRG_MANTISSA,-iTemp);
          test(); sub(pswChanEnrg[i], swTemp);
          if (pswChanEnrg[i] < swTemp)
          {
            pswChanEnrg[i] = swTemp;				move16();
          }
        }
      }
    }
  }
  else
  {
    for (i = 0; i < NUM_CHAN; i++)
    {
      pswChanNoiseEnrg[i] = MIN_CH_ENRG_MANTISSA;	move16();
    }
    iChanNoiseEnrgShift = MIN_CH_ENRG_SHIFT;		move16();
  }

}


/*
 * Compute the channel SNR indices
 */

for (i = 0; i < NUM_CHAN; i++)
{

  swTemp1 = pswChanEnrg[i];				move16();
  swTemp2 = pswChanNoiseEnrg[i];			move16();
  iShift = sub(iChanNoiseEnrgShift, iChanEnrgShift);

  test(); test();
  if ((swTemp1 == 0) || (swTemp2 == 0))
  {
    piChanSnr[i] = 0;					move16();
    continue;
  }

  test();
  if (sub(swTemp1, 0x4000) >= 0)
    {
    swTemp1 = shr(swTemp1, 1);
    iShift = add(iShift, 1);
    }

  test();
  while (sub(swTemp1,0x2000) < 0)
  {
    swTemp1 = shl(swTemp1,1);
    iShift = sub(iShift,1);
    test();
  }

  test();
  while (sub(swTemp1,swTemp2) >= 0)
  {
    swTemp2 = shl(swTemp2,1);
    iShift = add(iShift,1);
    test();
  }

  swLog2 = shl(fnLog2(L_deposit_h(div_s(swTemp1,swTemp2))),5);

  test();
  if (sub(swLog2,0x8000) <= 0)
  {
    swLog2 = 0;						move16();
  }
  else
  {
    swLog2 = add(swLog2,0x7FFF);
  }
  iShift = sub(iShift,1);

  test(); 
  if (sub(iShift,-1) <= 0)
  {
    piChanSnr[i] = 0;					move16();
    continue;
  }


/*
 * In the log2() domain (instead of 10*log10() domain),
 * 0.1875 = 2041, 0.375 = 4082, & 32768-8*4082 = 112
 */

  iChanSnr = shl(iShift,3);
  lwTemp = L_add(L_deposit_l(swLog2),shr(L_mult((Shortword)
                                                 iShift,(Shortword)112),1));
  test();
  if (L_sub(lwTemp,(Longword)2041) < 0)
  {
    iChanSnr = sub(iChanSnr,1);
    test();
    if (iChanSnr < 0)
    {
      iChanSnr = 0;					move16();
    }
    test(); sub(iChanSnr, 89);
    if (iChanSnr > 89)
    {
      iChanSnr = 89;					move16();
    }
    piChanSnr[i] = iChanSnr;				move16();
  }
  else
  {
    lwTemp = L_sub(lwTemp,(Longword)2041);
    test();
    while (L_sub(lwTemp,(Longword)4082) >= 0)
    {
      iChanSnr = add(iChanSnr,1);
      lwTemp = L_sub(lwTemp,(Longword)4082);
      test();
    }
    test(); 
    if (sub(iChanSnr, 89) >0 )
    {
      iChanSnr = 89;					move16();
    }
    piChanSnr[i] = iChanSnr;				move16();
  }

}


/*
* Compute the Voice Metric
*/

iVoiceMetric = 0;					move16();
for (i = 0; i < NUM_CHAN; i++)
{
  j = piChanSnr[i];					move16();
  iVoiceMetric = add(iVoiceMetric,piVMTable[j]);
}


/*
 * Estimate the log spectral energy deviation
 */

/*
 * First, compute the log-spectral energy
 */

for (i = 0; i < NUM_CHAN; i++)
{
  test();
  if (pswChanEnrg[i] == 0)
  {
    pnsLogSpecEnrg[i].iShift = 0;			move16();
    pnsLogSpecEnrg[i].swMantissa = 0;			move16();
    continue;
  }
  iShift = norm_s(pswChanEnrg[i]);
  swTemp = shl(pswChanEnrg[i],iShift);
  test();
  if (sub(swTemp,0x4000) == 0)
  {
    pnsLogSpecEnrg[i].iShift = add(add(iShift,iChanEnrgShift),1);
    							move16();
    pnsLogSpecEnrg[i].swMantissa = 0;			move16();
  }
  else
  {
    pnsLogSpecEnrg[i].iShift = add(iShift,iChanEnrgShift);	move16();
    pnsLogSpecEnrg[i].swMantissa = negate(shl(fnLog2(L_deposit_h(swTemp)),5));
    							move16();
  }
}

test(); test();
if ((iInitFramesFlag == TRUE) || (iFUpdateFlag == TRUE))
{
  for (i = 0; i < NUM_CHAN; i++)
  {
    pnsLogSpecEnrgLong[i].iShift = pnsLogSpecEnrg[i].iShift;	move16();
    pnsLogSpecEnrgLong[i].swMantissa = pnsLogSpecEnrg[i].swMantissa;
    							move16();
  }
}


/*
 * Next, compute the log-spectral energy deviation
 */

lwLogSpecEnrgDev = 0;					move16();
for (i = 0; i < NUM_CHAN; i++)
{
  lwTemp1 = L_add(L_deposit_h(pnsLogSpecEnrgLong[i].iShift),
            L_shl(L_deposit_l(pnsLogSpecEnrgLong[i].swMantissa),1));
  lwTemp2 = L_add(L_deposit_h(pnsLogSpecEnrg[i].iShift),
            L_shl(L_deposit_l(pnsLogSpecEnrg[i].swMantissa),1));
  lwLogSpecEnrgDev = L_add(lwLogSpecEnrgDev,L_abs(L_sub(lwTemp1,lwTemp2)));
}
test();
iLogSpecEnrgDevFlag = (L_sub(lwLogSpecEnrgDev,(Longword)DEV_THLD) >= 0);


/*
* Update the long term log-spectral energy
*/
test(); 
if (sub(iVoiceMetric, iSigThld) > 0)
{
  swGamma = HI_GAMMA;					move16();
  swGammaC = HI_GAMMA_COMPL;				move16();
}
else
{
  swGamma = LO_GAMMA;					move16();
  swGammaC = LO_GAMMA_COMPL;				move16();
}

for (i = 0; i < NUM_CHAN; i++)
{
  lwShift = L_mac(L_mult(swGamma,pnsLogSpecEnrgLong[i].iShift),swGammaC,pnsLogSpecEnrg[i].iShift);
  iShift = extract_h(lwShift);
  swMantissa = round(L_mac(L_mult(swGamma,pnsLogSpecEnrgLong[i].swMantissa),swGammaC,pnsLogSpecEnrg[i].swMantissa));
  lwMantissa = L_add((L_shr(lwShift,1) & 0x00007FFF),L_deposit_l(swMantissa)); logic32();

  test();
  if (L_sub(lwMantissa,0x00008000) >= 0)
  {
    pnsLogSpecEnrgLong[i].iShift = add(iShift, 1);	move16();
    pnsLogSpecEnrgLong[i].swMantissa = extract_l(lwMantissa) & 0x7FFF;
							move16(); logic16();
  }
  else
  {
    pnsLogSpecEnrgLong[i].iShift = iShift;		move16();
    pnsLogSpecEnrgLong[i].swMantissa = extract_l(lwMantissa);
							move16();
  }

}


/*
* Estimate the SNR of the speech input
*/

/*
* First, estimate the noise energy
*/

lwNoiseEnrg = 0;					move32();
for (i = 0; i < NUM_CHAN; i++)
{
  lwNoiseEnrg = L_add(lwNoiseEnrg,L_deposit_l(pswChanNoiseEnrg[i]));
}
iNoiseEnrgShift = iChanNoiseEnrgShift;			move16();


/*
* Next, estimate the signal energy
*/
test(); test();
if ((iInitFramesFlag == TRUE) || (iFUpdateFlag == TRUE))
{
  lwSigEnrgInst = INIT_SIG_ENRG_MANTISSA;		move32();
  iSigEnrgInstShift = INIT_SIG_ENRG_SHIFT;		move16();
}
else
{
  test(); 
  if (sub(iVoiceMetric, iSigThld) > 0)
  {
    lwSigEnrgInst = 0;					move32();
    for (i = 0; i < NUM_CHAN; i++)
    {
      test(); 
      if (sub(iChanNoiseEnrgShift, iChanEnrgShift) >= 0)
      {
        swTemp1 = pswChanEnrg[i];			move16();
        swTemp2 = shr(pswChanNoiseEnrg[i],
                         sub(iChanNoiseEnrgShift,iChanEnrgShift));
      }
      else
      {
        swTemp1 = shr(pswChanEnrg[i],
                         sub(iChanEnrgShift,iChanNoiseEnrgShift));
        swTemp2 = pswChanNoiseEnrg[i];			move16();
      }

      test();
      if (sub(swTemp1, swTemp2) > 0 )
      {
        lwSigEnrgInst = L_add(lwSigEnrgInst,L_deposit_l(swTemp1));
      }
      else
      {
        lwSigEnrgInst = L_add(lwSigEnrgInst,L_deposit_l(swTemp2));
      }
    }

    test();
    if (sub(iChanNoiseEnrgShift, iChanEnrgShift) >=0 )
    {
      iSigEnrgInstShift = iChanEnrgShift;		move16();
    }
    else
    {
      iSigEnrgInstShift = iChanNoiseEnrgShift;		move16();
    }
  }
  else
  {
    lwSigEnrgInst = lwNoiseEnrg;			move32();
    iSigEnrgInstShift = iNoiseEnrgShift;		move16();
  }
}


/*
* Compute the speech SNR
*/

test(); logic32();
if (lwSigEnrgInst & 0x00080000)
{
    swSigEnrgInst = extract_l(L_shr(lwSigEnrgInst,5));
    iSigEnrgInstShift = sub(iSigEnrgInstShift,5);
}
else 
{
    test(); logic32();
    if (lwSigEnrgInst & 0x00040000)
    {
        swSigEnrgInst = extract_l(L_shr(lwSigEnrgInst,4));
        iSigEnrgInstShift = sub(iSigEnrgInstShift,4);
    }
    else 
    {
        test(); logic32();
        if (lwSigEnrgInst & 0x00020000)
        {
            swSigEnrgInst = extract_l(L_shr(lwSigEnrgInst,3));
            iSigEnrgInstShift = sub(iSigEnrgInstShift,3);
        }
        else 
        {
            test(); logic32();
            if (lwSigEnrgInst & 0x00010000)
            {
                swSigEnrgInst = extract_l(L_shr(lwSigEnrgInst,2));
                iSigEnrgInstShift = sub(iSigEnrgInstShift,2);
            }
            else 
            {
                test(); logic32();
                if (lwSigEnrgInst & 0x00008000)
                {
                    swSigEnrgInst = extract_l(L_shr(lwSigEnrgInst,1));
                    iSigEnrgInstShift = sub(iSigEnrgInstShift,1);
                }
                else
                {
                    swSigEnrgInst = extract_l(lwSigEnrgInst);
                }
            }
        }
    }
}

test(); logic32();
if (lwNoiseEnrg & 0x00080000)
{
    swNoiseEnrg = extract_l(L_shr(lwNoiseEnrg,5));
    iNoiseEnrgShift = sub(iNoiseEnrgShift,5);
}
else 
{
    test(); logic32();
    if (lwNoiseEnrg & 0x00040000)
    {
        swNoiseEnrg = extract_l(L_shr(lwNoiseEnrg,4));
        iNoiseEnrgShift = sub(iNoiseEnrgShift,4);
    }
    else 
    {
        test(); logic32();
        if (lwNoiseEnrg & 0x00020000)
        {
            swNoiseEnrg = extract_l(L_shr(lwNoiseEnrg,3));
            iNoiseEnrgShift = sub(iNoiseEnrgShift,3);
        }
        else 
        {
            test(); logic32();
            if (lwNoiseEnrg & 0x00010000)
            {
                swNoiseEnrg = extract_l(L_shr(lwNoiseEnrg,2));
                iNoiseEnrgShift = sub(iNoiseEnrgShift,2);
            }
            else 
            {
                test(); logic32();
                if (lwNoiseEnrg & 0x00008000)
                {
                    swNoiseEnrg = extract_l(L_shr(lwNoiseEnrg,1));
                    iNoiseEnrgShift = sub(iNoiseEnrgShift,1);
                }
                else
                {
                    swNoiseEnrg = extract_l(lwNoiseEnrg);
                }
            }
        }
    }
}

test();
if (sub(swSigEnrgInst,0x4000) >= 0)
{
  swSigEnrgInst = shr(add(swSigEnrgInst,1),1);
  iSigEnrgInstShift = sub(iSigEnrgInstShift,1);
}
test();
while (sub(swSigEnrgInst,0x2000) < 0)
{
  swSigEnrgInst = shl(swSigEnrgInst,1);
  iSigEnrgInstShift = add(iSigEnrgInstShift,1);
  test();
}
test();
while (sub(swSigEnrgInst,swNoiseEnrg) >= 0)
{
  swNoiseEnrg = shl(swNoiseEnrg,1);
  iNoiseEnrgShift = add(iNoiseEnrgShift,1);
  test();
}

iLog2 = sub(iNoiseEnrgShift,iSigEnrgInstShift);
test();
if (swNoiseEnrg > 0)
{
  swLog2 = shl(fnLog2(L_deposit_h(div_s(swSigEnrgInst,swNoiseEnrg))),5);
}
else
{
  swLog2 = 0;					move16();
}
test();
if (sub(swLog2,0x8000) <= 0)
{
  swLog2 = 0;					move16();
}
else
{
  swLog2 = add(swLog2,0x7FFF);
}
iLog2 = sub(iLog2,1);

test(); 
if (sub(iLog2, 31) > 0)
{
  iLog2 = 31;					move16();
}
lwSnrInst = L_add(L_deposit_h(iLog2),L_shl(L_deposit_l(swLog2),1));
test();
if (lwSnrInst < 0)
{
  lwSnrInst = 0;				move32();
}
swSnrInst = extract_l(L_shr(lwSnrInst,6));

test(); test();
if ((iInitFramesFlag == TRUE) || (iFUpdateFlag == TRUE))
{
  swSnr = swSnrInst;				move16();
}
else
{
  test(); 
  if (sub(iVoiceMetric, iSigThld) >0)
  {
    swBetaC = negate(add(0x8000,swBeta));
    swSnr = round(L_mac(L_mult(swSnr,swBeta),swSnrInst,swBetaC));
    swBeta = add(swBeta,(Shortword)98);

    test(); 
    if (sub(swBeta, HI_BETA) >0)
    {
      swBeta = HI_BETA;				move16();
    }
  }
  else
  {
    swBeta = sub(swBeta,(Shortword)98);
    test(); 
    if (sub(swBeta, LO_BETA) <0)
    {
      swBeta = LO_BETA;				move16();
    }
  }
}


/*
* Quantize the SNR and select the different thresholds
* based on this value
*/

test(); test();
if (sub(swSnr, (Shortword)510) < 0)
{
  iQSnr = 0;					move16();
}
else if (sub(swSnr, (Shortword)9690) >= 0)
{
  iQSnr = 19;					move16();
}
else
{
  iQSnr = extract_h(L_mult(div_s(swSnr,(Shortword)9690),(Shortword)19));
}

iSigThld = piSigThld[iQSnr];			move16();
iUpdateThld = piUpdateThld[iQSnr];		move16();


/*
* Set or reset the update flag and the forced update flag
*/

iUpdateFlag = FALSE;				move16();
iFUpdateFlag = FALSE;				move16();

test(); test(); test();
test(); test();
if ((iInitFramesFlag == FALSE) && (sub(iVoiceMetric, iUpdateThld) < 0) &&
    (iPeak2AveFlag == FALSE))
{
  iUpdateFlag = TRUE;				move16();
  iUpdateCount = 0;				move16();
}
else if ((iPeak2AveFlag == FALSE) && (iLogSpecEnrgDevFlag == FALSE))
{
  iUpdateCount = add(iUpdateCount,1);
  test();
  if (sub(iUpdateCount,F_UPDATE_CNT_THLD) >= 0)
  {
    iUpdateFlag = TRUE;				move16();
    iFUpdateFlag = TRUE;			move16();
  }
}
else
{
  ;
}

test();
if (sub(iUpdateCount, iLastUpdateCount) == 0)
{
  iHysterCount = add(iHysterCount,1);
}
else
{
  iHysterCount = 0;				move16();
}
iLastUpdateCount = iUpdateCount;		move16();

test();
if (sub(iHysterCount,HYSTER_CNT_THLD) > 0)
{
  iUpdateCount = 0;				move16();
}


/*
* Update the channel noise estimates
*/
test();
if (iUpdateFlag == TRUE)
{

  test(); 
  if (sub(iChanNoiseEnrgShift, iChanEnrgShift) >=0 )
  {
    for (i = 0; i < NUM_CHAN; i++)
    {
      lwTemp = L_shr(L_mult(pswChanNoiseEnrg[i],(Shortword)
               CNE_SM_FAC_COMPL),sub(iChanNoiseEnrgShift,iChanEnrgShift));
      pswChanNoiseEnrg[i] = round(L_mac(lwTemp,(Shortword)CNE_SM_FAC,
                                     pswChanEnrg[i]));	move16();
    }
    iChanNoiseEnrgShift = iChanEnrgShift;	move16();
  }
  else
  {
    for (i = 0; i < NUM_CHAN; i++)
    {
      lwTemp = L_shr(L_mult(pswChanEnrg[i],(Shortword)CNE_SM_FAC),
                        sub(iChanEnrgShift,iChanNoiseEnrgShift));
      pswChanNoiseEnrg[i] = round(L_mac(lwTemp,(Shortword)
                                     CNE_SM_FAC_COMPL,pswChanNoiseEnrg[i]));
      						move16();
    }
  }

  swMax = pswChanNoiseEnrg[0];			move16();
  for (i = 1; i < NUM_CHAN; i++)
  {
    test();
    if (sub(swMax,pswChanNoiseEnrg[i]) < 0)
    {
      swMax = pswChanNoiseEnrg[i];		move16();
    }
  }
  iShift = norm_s(swMax);
  for (i = 0; i < NUM_CHAN; i++)
  {
    pswChanNoiseEnrg[i] = shl(pswChanNoiseEnrg[i],iShift); move16();
  }
  iChanNoiseEnrgShift = add(iChanNoiseEnrgShift,iShift);

  iTemp = sub(iChanNoiseEnrgShift,MIN_CH_ENRG_SHIFT);
  test();
  if (iTemp > 0)
  {
    for (i = 0; i < NUM_CHAN; i++)
    {
      pswChanNoiseEnrg[i] = (Shortword)MIN_CH_ENRG_MANTISSA;	move16();
    }
    iChanNoiseEnrgShift = MIN_CH_ENRG_SHIFT;		move16();
  }
  else
  {
    for (i = 0; i < NUM_CHAN; i++)
    {
      swTemp = shr(MIN_CH_ENRG_MANTISSA, -iTemp);
      test();
      if (sub(pswChanEnrg[i], swTemp) < 0)
      {
        pswChanEnrg[i] = swTemp;			move16();
      }
    }
  }
}


/*
* Save the output values and return
*/

*piQSnr = iQSnr;					move16();

test();
if (iInitFramesFlag == TRUE)
{
  iVoiceMetric = NON_SPEECH_THLD;			move16();
}

return(iVoiceMetric);

}

