/*=======================================================================================================*/
/*      Distributed Speech Recognition                                                                   */
/*      ETSI ES 202 050 Advanced Front-End Feature Extraction Algorithm & Compression Algorithm &        */
/*      ETSI ES 202 212 Extended Advanced Front-End Feature Extraction Algorithm & Compression Algorithm */  
/*      Bit exact C-language software implementation using ETSI basic operations library                 */
/*      3GPP TS 26.243 v 6.1.0 November 2004                                                                  */
/*      ANSI C code for the Fixed-Point Distributed Speech Recognition Extended Advanced Front-end       */
/*=======================================================================================================*/

#include  "ParmType.h"
#include "typedef.h"
#include "basic_op.h"
#include "count.h"

/*
 * Defines
 */

/* This function computes the square root of the input argument
 * using a 4 term Taylor expansions around 5/8 and 7/8.  The expansion
 * point is chosen to minimize error.
 */

#define FIX_34 (Shortword)(32768.0*3.0/4.0)
#define FIX_18 (Shortword)(32768.0*1.0/8.0)
#define FIX_INVSQRT2	(23170)			/* 1/sqrt(2) */


/*
 * Static variables
 */

   /* The following coefficients are scaled down by 2 to avoid saturating
    * the 32bit accumulator when arg approaches 1.
    */

   static const Shortword coeff_sqrt5_58[5]={
				   25905/2,	/*  sqrt(5/8)               */
				   20724/2,	/*  1/2*sqrt(8/5)           */
				  -8290/2,	/* -1/4*sqrt((8/5)^3)/2     */
				   6632/2,	/*  3/8*sqrt((8/5)^5)/6     */
				  -6632/2 };	/* -15/16*sqrt((8/5)^7)/24  */

   static const Shortword coeff_sqrt5_78[5]={
				   30652/2,	/*  sqrt(7/8)*0.5           */
				   17515/2,	/*  1/2*sqrt(8/7)*0.5       */
				  -5004/2,	/* -1/4*sqrt((8/7)^3)/2     */
				   2860/2,	/*  3/8*sqrt((8/7)^5)/6     */
				  -2043/2 };	/* -15/16*sqrt((8/7)^7)/24  */


/*
***********************************************************************
*
*   FUNCTION NAME: sqrt_l_fix
*
*   PURPOSE: To find the square root
*
*   INPUTS:
*
*      arg - Longword number to be square rooted
*
*   OUTPUTS:
*
*      None
*
*   RETURN VALUE:
*
*      The square root is returned
*
**********************************************************************
*/

/* Same as above except that input argument is a Longword (long) */

Shortword sqrt_l_fix(Longword L_arg)
   {
   Shortword x, y;
   Shortword k, l;
   const Shortword *coeff;
   Longword a;
   X_INT16 indx;

   test();
   if(L_arg==0)
      {
      return((Shortword)0);
      }
   k = norm_l(L_arg);		/* Get normalization factor. */
   x = round(L_shl(L_arg, k));		/* Normalize input. */
   x = sub(x,FIX_34);		/* x = norm(arg) - 3/4 */
   test();
   if(x>0)		/* norm(arg) > 3/4, use x0=7/8. */
      {
      x = sub(x,FIX_18);		/* x = norm(arg) - 7/8 */
      coeff = coeff_sqrt5_78; move16();
      }
   else				/* norm(arg) < 3/4, use x0=5/8. */
      {
      x = add(x,FIX_18);		/* x = norm(arg) - 5/8 */
      coeff = coeff_sqrt5_58; move16();
      }
   y = x; move16();
   a = L_deposit_h(coeff[0]);

   for(indx=1; indx<5; indx++)
      {
      a = L_mac(a, y, coeff[indx]);
      y = mult_r(x, y);
      }
   l = shr(k,1);
   a = L_shr(a, l-1);			/* Undo normalization */
   test();
   if(sub(k, shl(l,1))>0)
      {
      a = L_mult(round(a), FIX_INVSQRT2);	/* Compensate for 1/sqrt(2) (odd exponent). */
      }
   x = round(a);

   return(x);
   }



/***************************************************************************
 *
 *   FUNCTION NAME: fnLog2
 *
 *   PURPOSE: The purpose of this function is to take the log base 2 of input
 *            and divide by 32 and return; i.e. output = log2(input)/32
 *
 *   INPUTS:
 *
 *     Longword  L_Input
 *
 *   OUTPUTS:
 *
 *     None
 *
 *   RETURN VALUE:
 *
 *     Shortword  output
 *
 *   DESCRIPTION:
 *
 *     log2(x) = 4.0 * (-.3372223*x*x + .9981958*x -.6626105)
 *                           c0            c1          c2   (includes sign)
 *
 *   REFERENCE:  Sub-clause 4.1.8.2 of GSM Recommendation 06.20, eqn 3.9
 *
 *   KEYWORDS: log, logarithm, logbase2, fnLog2
 *
 *************************************************************************/

Shortword fnLog2(Longword L_Input)
{

/*_________________________________________________________________________
 |                                                                         |
 |                           Static Variables                              |
 |_________________________________________________________________________|
*/

  const static Shortword

         swC0 = -0x2b2a,
         swC1 = 0x7fc5,
         swC2 = -0x54d0;

/*_________________________________________________________________________
 |                                                                         |
 |                            Automatic Variables                          |
 |_________________________________________________________________________|
*/

  X_INT16 iShiftCnt;

  Shortword swInSqrd;
  Shortword swIn;


/*_________________________________________________________________________
 |                                                                         |
 |                              Executable Code                            |
 |_________________________________________________________________________|
*/

  /* normalize input and store shifts required */
  /* ----------------------------------------- */

  iShiftCnt = norm_l(L_Input);
  L_Input = L_shl(L_Input, iShiftCnt);
  iShiftCnt = add(iShiftCnt, 1);
  iShiftCnt = negate(iShiftCnt);

  /* calculate x*x*c0 */
  /* ---------------- */

  swIn = extract_h(L_Input);
  swInSqrd = mult_r(swIn, swIn);
  L_Input = L_mult(swInSqrd, swC0);

  /* add x*c1 */
  /* --------- */

  L_Input = L_mac(L_Input, swIn, swC1);

  /* add c2 */
  /* ------ */

  L_Input = L_add(L_Input, L_deposit_h(swC2));

  /* apply *(4/32) */
  /* ------------- */

  L_Input = L_shr(L_Input, 3);
  L_Input = L_Input & 0x03ffffff;               logic32();
  iShiftCnt = shl(iShiftCnt, 10);
  L_Input = L_add(L_Input, L_deposit_h(iShiftCnt));

  /* return log */
  /* ---------- */

  return (round(L_Input));
}


