/*=======================================================================================================*/
/*      Distributed Speech Recognition                                                                   */
/*      ETSI ES 202 050 Advanced Front-End Feature Extraction Algorithm & Compression Algorithm &        */
/*      ETSI ES 202 212 Extended Advanced Front-End Feature Extraction Algorithm & Compression Algorithm */  
/*      Bit exact C-language software implementation using ETSI basic operations library                 */
/*      3GPP TS 26.243 v 6.1.0 November 2004                                                                  */
/*      ANSI C code for the Fixed-Point Distributed Speech Recognition Extended Advanced Front-end       */
/*=======================================================================================================*/
/*---------------------------------------------------------------------------
 *
 * FILE NAME: ParmInterface_B.h
 * PURPOSE:   Function prototypes for ParmInterface_B.c
 *
 *---------------------------------------------------------------------------*/
#ifndef _PARMINTERFACE_B_H
#define _PARMINTERFACE_B_H

/*-----------------
 * File Inclusions
 *-----------------*/
#include "BufferIn.h"
#include "ParmType.h"
#include "BufferIn32.h"

/*------------------------------------
 * Definition of Front-end Parameters 
 *------------------------------------*/
#define SAMPLING_FREQ_1    8   // 8kHz
#define SAMPLING_FREQ_2   11   // 11kHz
#define SAMPLING_FREQ_3   16   // 16kHz

#define FRAME_LENGTH     200   // 25ms

#define FRAME_SHIFT       80   // 10ms

#define FFT_LENGTH       256
#define FRAME_BUF_SIZE   241

#define STARTING_FREQ     64.0 // 55.401825

#define NUM_CEP_COEFF     13   //c1...c12 + c0

#define NON_SPEECH_FRAME   0
#define SPEECH_FRAME       1

#define PIx2               6.28318530717958647692

/*------------------------------------------------------
 * Private structures for the modules (noise reduction,
 * SNR waveform processing, cepstrum calculation, ...)
 *------------------------------------------------------*/
typedef struct NoiseSupStructX NoiseSupStructX;
typedef struct WaveProcStructX WaveProcStructX;
typedef struct CompCepsStructX CompCepsStructX;
typedef struct PostProcStructX PostProcStructX;
typedef struct DataFor16kProc  DataFor16kProc;
typedef struct MelFB_Window    MelFB_Window;
typedef struct VADStructX      VADStructX;



#define SIGNAL_TYPE float
#define COEF_TYPE   float


#define ROUND16(a)      ((X_INT16)( ((a) >= 0) ? ((X_INT16)((a) + 0.5)) : ((X_INT16)((a) - 0.5))))
#define ROUND32(a)      ( ((a) >= 0) ? ((X_INT32)((a) + 0.5)) : ((X_INT32)((a) - 0.5)))


#define MAX_NOISE_SHFT (FRAME_LENGTH/FRAME_SHIFT+2)


/* mutriplicative factor for arguments */
#define PP_FACT_COEF    256
#define PP_FACT_ENER    64
#define PP_FACT_C0              32

/*------------------------------------------------------
 * FEParamsX must (only) contains:
 *  - initialisation parameters
 *    (like FrameLength, FrameShift, FFTLength)
 *  - buffers to transmit data between modules
 *    (like denoisedBuf)
 *  - specific information which must be remanent
 *    between two frames, and which can't be 
 *    encapsulated in modules
 *  - if a data is not remanent, they must be transmitted
 *    between modules by a GET ans a SET function
 *  - if a data can be encapsulated in a module, it must
 *    be encapsulated to avoid conflict between modules
 *------------------------------------------------------*/
typedef struct FEParamsX_F FEParamsX_F;


struct FEParamsX_F
{
    BOOLEAN Noc0;
        
    int VAD;
    int CoefNb;
    int FFTLength;
    int FrameShift;
    int FrameLength;
    int FrameCounter;
    int SpeechFoundMel;
    int SpeechFoundVar;
    int SpeechFoundSpec;
    int SpeechFoundVADNS;
    int NbSamplesToRead;
    int SamplingFrequency;
    int offsetDenoisedFrame;
        
    float StartingFrequency;
        
    long NonZeroFrameOnset;
    long ZeroFrameCounter;
        
    X_INT16 tabShiftNoiseToCep[MAX_NOISE_SHFT];   
    BufferIn32 *denoisedBuf32;
#if defined(F_16KHZ)
    BOOLEAN Do16kHzProc;
    struct DataFor16kProc_F *pData16k;
#endif  


/*
 * For XAFE
 */
    FILE *pitch_fid;
    FILE *class_fid;
    void *pRvcPitchRom;
    void *pRvcPitchMeter;
    /*---------------------------
     * module specific expansion
     *---------------------------*/
    struct NoiseSupStructX_F *NSX;
    WaveProcStructX *WPX;
    CompCepsStructX *CCX;
    PostProcStructX *PPX;
    struct VADStructX_F *VADX;
        
    /*-------------------------------------
     * methods: use pointer to function to
     *          emulate generic function
     *-------------------------------------*/
        
    /*-----------------------
     * for noise suppression 
     *-----------------------*/
    struct NoiseSupStructX_F * (*DoNoiseSupAlloc) (void);
    void (*DoNoiseSupInit) (FEParamsX_F *this);

#if defined(XAFE)
    BOOLEAN (*DoNoiseSup) (X_INT16 *InData16,
                           X_INT32 *OutData32,
                           FEParamsX_F *this,
                           X_INT16 *shftOut,
                           Shortword *pswStft,
                           Shortword *pswSignal,
                           X_INT16 *piStftShift,
                           X_INT16 *pidecalSig,
                           X_INT16 *piProcessFlag);
#else
    BOOLEAN (*DoNoiseSup) (X_INT16 *InData16,
                           X_INT32 *OutData32,
                           FEParamsX_F *this,
                           X_INT16 *shftOut);
#endif
    void (*DoNoiseSupDelete) (struct NoiseSupStructX_F *);
        
    /*-------------------------
     * for waveform processing
     *-------------------------*/
    WaveProcStructX* (*DoWaveProcAlloc) (void);
    void (*DoWaveProcInit) (FEParamsX_F *this);
    BOOLEAN (*DoWaveProc) (X_INT32 *DataInSh32,
                           X_INT16 shiftIn,
                           X_INT16 *dataOut,X_INT16 *shiftOut,FEParamsX_F *this);
    void (*DoWaveProcDelete) (WaveProcStructX*);
        
    /*--------------------------
     * for cepstrum calculation
     *--------------------------*/
    CompCepsStructX* (*DoCompCepsAlloc) (void);
    void (*DoCompCepsInit) (FEParamsX_F *this);
    BOOLEAN (*DoCompCeps) (X_INT16 *Data,X_INT16 shiftIn, 
                           X_INT16 *Coef, 
                           FEParamsX_F *this);
    void (*DoCompCepsDelete) (CompCepsStructX*);
        
    /*---------------------
     * for post processing 
     *---------------------*/
    PostProcStructX* (*DoPostProcAlloc) (void);
    void (*DoPostProcInit) (FEParamsX_F *this);
    BOOLEAN (*DoPostProc) (X_INT16 *Coef, 
                           FEParamsX_F *this);
    void (*DoPostProcDelete) (PostProcStructX *);
        
    /*----------------------
     * for 16kHz Processing
     *----------------------*/
#if defined(F_16KHZ)
    struct DataFor16kProc_F *(*Do16kProcAlloc) (void);
    void (*Do16kProcInit) (FEParamsX_F * This);
    void (*Do16kProcessing) (float *inData, 
                             struct DataFor16kProc_F *pData16k, 
                             int inDataLength);
    void (*Do16kProcDelete) (struct DataFor16kProc_F * pData16k);
#endif  
    /*---------
     * for VAD
     *---------*/
    struct VADStructX_F *(*DoVADAlloc) (void);
    void (*DoVADInit) (FEParamsX_F * This);
    BOOLEAN (*DoVADProc) (X_INT16 *FeatureBuffer,FEParamsX_F * This);
    void (*DoVADDelete) (struct VADStructX_F *VADX);
        
    /*-------------------
     * for miscellaneous 
     *-------------------*/
    BOOLEAN (*DoVADFlush) (X_INT16 *FeatureBuffer,FEParamsX_F * This);

};

/*----------------------------------------------------
 * Global routines : each call the four stages of the
 * front end (noise suppression, wave processing,
 * cepstrum calculation and post processing), and
 * finally call the VAD stage
 *----------------------------------------------------*/

/*------------
 * allocation
 *------------*/
extern FEParamsX_F* AdvProcessAlloc_B (int samplingFrequency);

/*----------------
 * initialisation
 *----------------*/
extern void AdvProcessInit_B (FEParamsX_F *pComParX);

/*-------------------------------------------------------------
 * Processing "FrameShift" samples
 * Return TRUE if a frame is output, FALSE if no (because some
 * modules may have some latency, hence at the beginning there
 * is no output data for the first input frames
 *-------------------------------------------------------------*/
extern BOOLEAN DoAdvProcess_B (FILE_TYPE *CurrentFrame, 
                               COEF_TYPE *FeatureBuffer,
                               FEParamsX_F *pComParX);

/*-------------------------------------------------------------
 * Flushing: (at the end of a record).
 * Return TRUE if a frame is flushed, 
 * FALSE if there is no more frame.
 *-------------------------------------------------------------*/
extern BOOLEAN FlushAdvProcess_B (COEF_TYPE *FeatureBuffer,
                                  FEParamsX_F *pComParX);

/*----------
 * deletion
 *----------*/
extern void AdvProcessDelete_B (FEParamsX_F **);

#endif
